"""
SUDO Responder Tests.

:requirement: sudo
"""

from __future__ import annotations

import re
import time
from datetime import datetime, timedelta

import pytest
from sssd_test_framework.roles.ad import AD
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericADProvider, GenericProvider
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.roles.samba import Samba
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_sudo__user_allowed(client: Client, provider: GenericProvider, sssd_service_user: str):
    """
    :title: One user is allowed to run command, other user is not
    :setup:
        1. Create user "user-1"
        2. Create user "user-2"
        3. Create sudorule to allow "user-1" run "/bin/ls on all hosts
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/ls root" as user-1
        3. List sudo rules for "user-2"
        4. Run "sudo /bin/ls root" as user-2
    :expectedresults:
        1. User is able to run /bin/ls as root
        2. Command is successful
        3. User is not able to run /bin/ls as root
        4. Command failed
    :customerscenario: False
    """
    u = provider.user("user-1").add()
    provider.user("user-2").add()
    provider.sudorule("test").add(user=u, host="ALL", command="/bin/ls")

    client.sssd.common.sudo()
    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(root) /bin/ls"]), "Sudo list failed!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/ls /root"), "Sudo command failed!"

    assert not client.auth.sudo.list("user-2", "Secret123"), "Sudo list successful!"
    assert not client.auth.sudo.run("user-2", "Secret123", command="/bin/ls /root"), "Sudo command successful!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.topology(KnownTopology.Samba)
def test_sudo__duplicate_sudo_user(client: Client, provider: GenericProvider):
    """
    :title: User is mentioned twice in sudoUser attribute, once with shortname and once with fully qualified name
    :setup:
        1. Create users "user-1", "user-2", "user-3", "user-4"
        3. Create sudorule to allow "user-1", "user-2", "user-2@test", "user-3" run "/bin/ls on all hosts
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1", "user-2", "user-3"
        2. Run "sudo /bin/ls root" as "user-1", "user-2", "user-3"
        3. List sudo rules for "user-4"
        4. Run "sudo /bin/ls root" as "user-4"
    :expectedresults:
        1. User is able to run /bin/ls as root
        2. Command is successful
        3. User is not able to run /bin/ls as root
        4. Command failed
    :customerscenario: False

    Note: This test can not run on IPA since it will not allow this case to happen.
    """
    provider.user("user-1").add()
    provider.user("user-2").add()
    provider.user("user-3").add()
    provider.user("user-4").add()
    provider.sudorule("test").add(
        user=["user-1", "user-2", f"user-2@{client.sssd.default_domain}", "user-3"], host="ALL", command="/bin/ls"
    )

    client.sssd.common.sudo()
    client.sssd.start()

    # Try several users to make sure we don't mangle the list
    for user in ["user-1", "user-2", "user-3"]:
        assert client.auth.sudo.list(user, "Secret123", expected=["(root) /bin/ls"]), "Sudo list failed!"
        assert client.auth.sudo.run(user, "Secret123", command="/bin/ls /root"), "sudo command failed!"

    assert not client.auth.sudo.list("user-4", "Secret123"), "Sudo list successful!"
    assert not client.auth.sudo.run("user-4", "Secret123", command="/bin/ls /root"), "Sudo command successful!"


@pytest.mark.importance("critical")
@pytest.mark.ticket(bz=1380436, gh=4236)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__case_sensitive_false(client: Client, provider: GenericProvider):
    """
    :title: Sudo rules work correctly for case-insensitive domains
    :setup:
        1. Create user "USER-1"
        2. Create sudorule to allow "user-1" run "/bin/less on all hosts
        3. Create sudorule to allow "USER-1" run "/bin/more on all hosts
        4. Enable SSSD sudo responder
        5. Set "case_sensitive" to "false"
        6. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/less root" as user-1
        3. Run "sudo /bin/more root" as user-1
        4. List sudo rules for "USER-1"
        5. Run "sudo /bin/less root" as USER-1
        6. Run "sudo /bin/more root" as USER-1
    :expectedresults:
        1. User is able to run /bin/less and /bin/more as root
        2. Command is successful
        3. Command is successful
        4. User is able to run /bin/less and /bin/more as root
        5. Command is successful
        6. Command is successful
    :customerscenario: False
    """
    provider.user("USER-1").add()
    provider.sudorule("lowercase").add(user="user-1", host="ALL", command="/bin/less")
    provider.sudorule("uppsercase").add(user="USER-1", host="ALL", command="/bin/more")
    client.fs.write("/root/test", "test")

    client.sssd.common.sudo()
    client.sssd.domain["case_sensitive"] = "false"
    client.sssd.start()

    assert client.auth.sudo.list(
        "user-1", "Secret123", expected=["(root) /bin/less", "(root) /bin/more"]
    ), "Sudo list failed!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/less /root/test"), "Sudo command failed!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/more /root/test"), "Sudo command failed!"

    assert client.auth.sudo.list(
        "USER-1", "Secret123", expected=["(root) /bin/less", "(root) /bin/more"]
    ), "Sudo list failed!"
    assert client.auth.sudo.run("USER-1", "Secret123", command="/bin/less /root/test"), "Sudo command failed!"
    assert client.auth.sudo.run("USER-1", "Secret123", command="/bin/more /root/test"), "Sudo command failed!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_sudo__rules_refresh(client: Client, provider: GenericProvider, sssd_service_user: str):
    """
    :title: Sudo rules refresh works
    :setup:
        1. Create user "user-1"
        2. Create sudorule to allow "user-1" run "/bin/ls on all hosts
        3. Enable SSSD sudo responder
        4. Set "entry_cache_sudo_timeout" to "2"
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Modify the rule to allow only "/bin/less" command
        3. Wait until the cached rule is expired (3 seconds)
        4. List sudo rules for "user-1"
    :expectedresults:
        1. User is able to run only /bin/ls
        2. Rule was modified
        3. Time passed
        4. User is able to run only /bin/less
    :customerscenario: False
    """
    u = provider.user("user-1").add()
    r = provider.sudorule("test").add(user=u, host="ALL", command="/bin/ls")

    client.sssd.common.sudo()
    client.sssd.domain["entry_cache_sudo_timeout"] = "2"
    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(root) /bin/ls"]), "Sudo list failed!"
    r.modify(command="/bin/less")
    time.sleep(3)
    assert client.auth.sudo.list("user-1", "Secret123", expected=["(root) /bin/less"]), "Sudo command failed!"


@pytest.mark.importance("critical")
@pytest.mark.ticket(bz=1372440, gh=4236)
@pytest.mark.contains_workaround_for(gh=4483)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__user_is_group(client: Client, provider: GenericProvider):
    """
    :title: POSIX groups can be set in sudoUser attribute
    :setup:
        1. Create user "user-1"
        2. Create group "group-1" with "user-1" as a member
        3. Create sudorule to allow "group-1" run "/bin/ls on all hosts
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/ls" as "user-1"
    :expectedresults:
        1. User is able to run only /bin/ls
        2. Command is successful
    :customerscenario: False
    """
    u = provider.user("user-1").add()
    g = provider.group("group-1").add().add_member(u)
    provider.sudorule("test").add(user=g, host="ALL", command="/bin/ls")

    client.sssd.common.sudo()
    client.sssd.start()

    # Until https://github.com/SSSD/sssd/issues/4483 is resolved
    # Running 'id user-1' will resolve SIDs into group names
    if isinstance(provider, (AD, Samba)):
        client.tools.id("user-1")

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(root) /bin/ls"]), "Sudo list failed!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/ls /root"), "Sudo command failed!"


@pytest.mark.importance("critical")
@pytest.mark.ticket(bz=1826272, gh=5119)
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_sudo__user_is_nonposix_group(client: Client, provider: GenericADProvider):
    """
    :title: Non-POSIX groups can be set in sudoUser attribute
    :setup:
        1. Create user "user-1"
        2. Create group "group-1" with "user-1" as a member
        3. Create sudorule to allow "group-1" run "/bin/ls on all hosts
        4. Enable SSSD sudo responder
        5. Disable ldap_id_mapping
        6. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/ls" as "user-1"
    :expectedresults:
        1. User is able to run only /bin/ls
        2. Command is successful
    :customerscenario: False
    """
    u = provider.user("user-1").add(uid=10001, gid=10001)
    g = provider.group("group-1").add().add_member(u)
    provider.sudorule("test").add(user=g, host="ALL", command="/bin/ls")

    client.sssd.common.sudo()
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(root) /bin/ls"]), "Sudo list failed!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/ls /root"), "Sudo command failed!"


@pytest.mark.importance("critical")
@pytest.mark.ticket(bz=1910131)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__runasuser_shortname(client: Client, provider: GenericADProvider):
    """
    :title: sudoRunAsUser contains shortname
    :setup:
        1. Create user "user-1"
        2. Create user "user-2"
        3. Create sudorule to allow "user-1" run "/bin/ls on all hosts as "user-2" using shortname
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
    :expectedresults:
        1. User is able to run /bin/ls as "user-2"
    :customerscenario: True
    """
    u1 = provider.user("user-1").add()
    provider.user("user-2").add()
    provider.sudorule("test").add(user=u1, host="ALL", command="/bin/ls", runasuser="user-2")

    client.sssd.common.sudo()
    client.sssd.start()

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(user-2) /bin/ls"]), "Sudo list failed!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.topology(KnownTopology.Samba)
def test_sudo__runasuser_fqn(client: Client, provider: GenericProvider):
    """
    :title: sudoRunAsUser contains fully qualified name
    :setup:
        1. Create user "user-1"
        2. Create user "user-2"
        3. Create sudorule to allow "user-1" run "/bin/ls on all hosts as "user-2" using fully qualified name
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
    :expectedresults:
        1. User is able to run /bin/ls as "user-2"
    :customerscenario: False

    Note: This test can not run on IPA since it does not allow fully qualified name here.
    """
    u1 = provider.user("user-1").add()
    provider.user("user-2").add()
    provider.sudorule("test").add(
        user=u1, host="ALL", command="/bin/ls", runasuser=f"user-2@{client.sssd.default_domain}"
    )

    client.sssd.common.sudo()
    client.sssd.start()

    assert client.auth.sudo.list("user-1", "Secret123", expected=["(user-2) /bin/ls"]), "Sudo list failed!"


@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_sudo__sudonotbefore_shorttime(client: Client, provider: LDAP):
    """
    :title: sudoNotBefore and sudoNotAfter do not require minutes and seconds
    :description: Test that sudoNotBefore and sudoNotAfter works even without minutes and seconds specifier.
    :setup:
        1. Create user "user-1"
        2. Create sudorule to allow "user-1" run "/bin/ls on all hosts within given time in %Y%m%d%H format
        3. Enable SSSD sudo responder
        4. Set "sudo_timed" to "true"
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
    :expectedresults:
        1. User is able to run /bin/ls within given time
    :customerscenario: False

    Note: IPA does not support these attributes and AD/Samba time schema
    requires minutes and seconds to be set. Therefor this test only applies to
    LDAP.
    """

    def shorttime(t: datetime) -> str:
        return t.strftime("%Y%m%d%H") + "Z"

    def fulltime(t: datetime) -> str:
        return t.strftime("%Y%m%d%H%M%S") + "Z"

    now = datetime.today().replace(minute=0, second=0, microsecond=0)
    notbefore = now - timedelta(days=1)
    notafter = now + timedelta(days=1)

    u = provider.user("user-1").add()
    provider.sudorule("test").add(
        user=u,
        host="ALL",
        command="/bin/ls",
        notbefore=shorttime(datetime.today() - timedelta(days=1)),
        notafter=shorttime(datetime.today() + timedelta(days=1)),
    )

    client.sssd.common.sudo()
    client.sssd.sudo["sudo_timed"] = "true"
    client.sssd.start()

    assert client.auth.sudo.list(
        "user-1",
        "Secret123",
        expected=[f"(root) NOTBEFORE={fulltime(notbefore)} NOTAFTER={fulltime(notafter)} /bin/ls"],
    ), "Sudo list failed!"


# This test is testing randomized values, therefore it is possible that
# the same refresh interval is generated multiple times (i.e. full refresh
# is always 4 seconds) and therefore the test fails. However, we want to check
# that random values are assigned so we must repeat the test in this case.
@pytest.mark.flaky(max_runs=5)
@pytest.mark.importance("low")
@pytest.mark.slow(seconds=15)
@pytest.mark.ticket(bz=1925514, gh=5609)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__refresh_random_offset(client: Client):
    """
    :title: Random offset is applied to sudo full and smart refresh.
    :setup:
        1. Set ldap_sudo_full_refresh_interval to 2
        2. Set ldap_sudo_smart_refresh_interval to 1
        3. Set ldap_sudo_random_offset to 5
        4. Start SSSD
    :steps:
        1. Grep domain log to see when the full refresh was scheduled
        2. Grep domain log to see when the smart refresh was scheduled
    :expectedresults:
        1. It was scheduled to multiple random times
        2. It was scheduled to multiple random times
    :customerscenario: True
    """
    client.sssd.domain.update(
        ldap_sudo_full_refresh_interval="2",
        ldap_sudo_smart_refresh_interval="1",
        ldap_sudo_random_offset="5",
    )

    # Start SSSD and wait for few sudo updates to occur
    client.sssd.start()
    time.sleep(15)

    # Stop SSSD to ensure that all logs are written
    client.sssd.stop()

    log = client.fs.read(client.sssd.logs.domain())
    smart = set()
    full = set()
    for m in re.findall(r"Task \[SUDO (Smart|Full) Refresh\]: scheduling task (\d+) seconds", log):
        match m[0]:
            case "Smart":
                smart.add(m[1])
            case "Full":
                full.add(m[1])

    assert len(smart) > 1, "Smart refresh scheduling tasks is > 1!"
    assert len(full) > 1, "Full refresh scheduling tasks is > 1!"


@pytest.mark.importance("low")
@pytest.mark.slow(seconds=10)
@pytest.mark.ticket(bz=1925505, gh=5604)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize(["full_interval", "smart_interval"], [(2, 1), (3, 2)])
def test_sudo__prefer_full_refresh_over_smart_refresh(client: Client, full_interval: int, smart_interval: int):
    """
    :title: Sudo smart refresh does not occur at the same time as full refresh
    :setup:
        1. Set ldap_sudo_full_refresh_interval to @full_interval
        2. Set ldap_sudo_smart_refresh_interval to @smart_interval
        3. Set ldap_sudo_random_offset to 0
        4. Start SSSD
    :steps:
        1. Sleep for 10 seconds
        2. Grep domain log to see when the smart and full refresh happened
    :expectedresults:
        1. Time passed
        2. Smart refresh does not occur at the same time as full refresh
    :customerscenario: True
    """

    def is_task_start(task: str, line: str) -> bool:
        return f"Task [{task}]: executing task" in line

    def is_task_end(task: str, line: str) -> bool:
        return f"Task [{task}]: finished successfully" in line or f"Task [{task}]: failed" in line

    def is_smart_skipped(line: str) -> bool:
        return "Skipping smart refresh because there is ongoing full refresh." in line

    client.sssd.domain.update(
        ldap_sudo_full_refresh_interval=str(full_interval),
        ldap_sudo_smart_refresh_interval=str(smart_interval),
        ldap_sudo_random_offset="0",
    )
    client.sssd.start()
    time.sleep(10)
    log = client.fs.read(client.sssd.logs.domain())

    expect_skip = False
    inside_full = False
    is_skipped = False

    # Check that
    # - Either there is no smart refresh executed inside a full refresh
    # - Or the smart refresh was skipped
    for line in log.splitlines():
        if is_task_start("SUDO Full Refresh", line):
            inside_full = True

        if is_task_end("SUDO Full Refresh", line):
            inside_full = False

        if is_task_start("SUDO Smart Refresh", line):
            is_skipped = False
            if inside_full:
                expect_skip = True

        if is_task_end("SUDO Smart Refresh", line):
            assert not expect_skip or is_skipped, "Sudo refresh was not skipped!"
            is_skipped = False
            expect_skip = False

        if is_smart_skipped(line):
            is_skipped = True


@pytest.mark.importance("high")
@pytest.mark.ticket(bz=1294670, gh=3969)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_sudo__local_users_negative_cache(client: Client, provider: LDAP, sssd_service_user: str):
    """
    :title: Sudo responder hits negative cache for local users
    :setup:
        1. Create local user "user-1"
        2. Add local rule to /etc/sudoers to allow all commands for "user-1"
        3. Enable sudo responder
        4. Set entry_negative_timeout to 0 to disable standard negative cache
        5. Start SSSD
    :steps:
        1. Authenticate as "user-1" over SSH
        2. Run "sudo /bin/ls /root"
        3. Start tcpdump to capture ldap packets and run "sudo /bin/ls /root" multiple times again
    :expectedresults:
        1. User is logged into the host
        2. Command is successful, user is stored in negative cache for local users
        3. No ldap packets for "user-1" user resolution are sent
    :customerscenario: True

    First sudo goes through SSSD to lookup up the user in LDAP, since it is not
    there and the user is local, it is stored in negative cache with very long
    expiration time. Subsequent sudo requests will hit the negative cache and no
    further lookup is performed.
    """
    client.local.user("user-1").add()
    client.fs.write("/etc/sudoers.d/test", "user-1 ALL=(ALL) NOPASSWD:ALL")

    client.sssd.common.sudo()
    client.sssd.nss.update(
        entry_negative_timeout="0",  # disable standard negative cache to make sure we hit the local user case
    )
    client.sssd.start(service_user=sssd_service_user)

    # Now there should be no query
    with client.ssh("user-1", "Secret123") as ssh:
        ssh.exec(["sudo", "/bin/ls", "/root"])

        with client.tools.tcpdump("/tmp/sssd.pcap", ["-s0", "host", provider.host.hostname]):
            ssh.exec(["sudo", "/bin/ls", "/root"])
            ssh.exec(["sudo", "/bin/ls", "/root"])

    result = client.tools.tshark(["-r", "/tmp/sssd.pcap", "-V", "-2", "-R", "ldap.filter"])
    assert "uid=user-1" not in result.stdout, "Packets sent when resolving user!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__defaults_set_no_auth_and_sudorule_has_auth_undefined(client: Client, provider: GenericProvider):
    """
    :title: Defaults sudo entry set !authentication and a sudo rule with undefined authentication
    :setup:
        1. Create user "user-1"
        2. Create an entry in cn=sudoers container defaults with option '!authenticate'
        3. Create a sudorule named allow-user-1 to allow user-1 to run all commands on all hosts, keep
           authentication option unset
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/ls root" as user-1
    :expectedresults:
        1. User is able to run sudo commands on client with password authentication
        2. Command is successful without password authentication
    :customerscenario: False
    """
    provider.user("user-1").add()
    provider.sudorule("defaults").add(option="!authenticate")
    provider.sudorule("allow-user-1").add(user="user-1", host="ALL", command="ALL")

    client.sssd.common.sudo()
    client.sssd.start()

    assert client.auth.sudo.list("user-1", expected=["(root) ALL"]), "Sudo list failed!"
    assert client.auth.sudo.run("user-1", command="/bin/ls /root"), "Sudo command failed!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_sudo__defaults_set_no_auth_and_sudo_rule_has_mandatory_auth(client: Client, provider: GenericProvider):
    """
    :title: Defaults sudo entry set !authentication and a sudorule has mandatory authentication
    :setup:
        1. Create user "user-1"
        2. Create a sudorule named defaults with option '!authenticate'
        3. Create a sudorule named allow-user-1 to allow user-1 to run all commands on all hosts with authentication
           forced
        4. Enable SSSD sudo responder
        5. Start SSSD
    :steps:
        1. List sudo rules for "user-1"
        2. Run "sudo /bin/ls root" as user-1
    :expectedresults:
        1. User should be able to run sudo commands on client with mandatory password authentication
        2. Command is successful only with password authentication
    :customerscenario: False
    """
    provider.user("user-1").add()
    provider.sudorule("defaults").add(option="!authenticate")
    provider.sudorule("allow-user-1").add(user="user-1", host="ALL", command="ALL", option="authenticate")

    client.sssd.common.sudo()
    client.sssd.start()

    assert client.auth.sudo.list("user-1", expected=["(root) PASSWD: ALL"]), "Sudo list failed!"
    assert not client.auth.sudo.run("user-1", command="/bin/ls /root"), "Sudo command successful!"
    assert client.auth.sudo.run("user-1", "Secret123", command="/bin/ls /root"), "Sudo command failed!"



"""Automation for krb ldap connection

:requirement: krb_ldap_connection
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import subprocess
import time
import pytest
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name


@pytest.mark.krbldapconnection
@pytest.mark.tier1_3
@pytest.mark.usefixtures('setup_sssd',
                         'create_posix_usersgroups',
                         'krb_connection_timeout')
class TestKrbLdapConnectionTimeout(object):
    """
    This is test case class for krb_ldap_connection suite

    Test connection expiration between SSSD and LDAP server. Test for
    default connection expire timeout and test for some arbitrary time period,
    (in our case, 100 second), invalid timeout period (-100 second), and 0
    second.

    timeouts that we are testing for. 'default' implies default timeout for
    sssd which is 900 seconds/15 min. 'timeout_out_of_range' is a timeout
    beyond the integer range. Also test the connection timeout between SSSD
    and Kerberos server. So, sssd won't restart succesfully. 'krb' implies
    connection expires when ticket expires (2 min for our test case).
    Connection expires as soon as TGT expires (2 min in our case).
    """
    def test_0001_timeoutdefault(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test if connection expires for the default value of ldap connection
         timeout that is 900 seconds(15 minutes) after that it should release
         the connection
        :id: 53ba0b29-f5fc-4daa-8730-04a8aec91829
        """
        domain_params = {'ldap_connection_expire_timeout': None}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params, 'delete')
        multihost.client[0].log.info(
            '\n\n\nTesting for default value of ldap_'
            'connection_expire_timeout; i.e. ldap_connection_'
            'expire_timeout = default')
        domain_params = {
            'ldap_uri': 'ldap://%s' % (multihost.master[0].sys_hostname)}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        self.run_test(900, multihost)

    def test_0002_timeout100(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test for arbitrary value ldap connection timeout that is 100
         seconds after that it should release the connection
        :id: bb8dee0a-8ade-4618-b616-589bfcd46ef3
        """
        multihost.client[0].log.info(
            '\n\n\nTesting for ldap_connection_expire_'
            'timeout = 100')
        domain_params = {
            'ldap_uri': 'ldap://%s' % (multihost.master[0].sys_hostname)}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        domain_params = {'ldap_connection_expire_timeout': 100}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)

        self.run_test(100, multihost)

    def test_0003_timeouttimeoutoutofrange(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test for out of range value of ldap connection timeout that
         is value out of range of integer
        :id: a3773739-41c7-4379-82d0-721d6993633c
        :expectedresults: SSSD sevice must fail to restart
         successfully after entering that value in configuration
        """

        multihost.client[0].log.info(
            '\n\n\nTesting for the case where timeout value is'
            'out of range (of integer). ldap_connection_expire_'
            'timeout = timeout_out_of_range')
        cmd_max_value = "echo $((`getconf INT_MAX`+1))"
        cmd = multihost.client[0].run_command(cmd_max_value)
        timeout = int(cmd.stdout_text.replace("\n", ""))
        domain_params = {
            'ldap_uri': 'ldap://%s' % (multihost.master[0].sys_hostname)}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        domain_params = {'ldap_connection_expire_timeout': timeout}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        tools = sssdTools(multihost.client[0])
        domainname = tools.get_domain_section_name()
        tools.delete_sssd_domain_log(domainname)
        logfile = '/var/log/sssd/sssd_%s.log' % ds_instance_name

        clear_sssd_cache = True
        try:
            # stop sssd service
            multihost.client[0].service_sssd('stop')
            # remove sssd cache
            location = '/var/lib/sss/db/'
            if not sssdTools(multihost.client[0]).remove_sss_cache(location):
                multihost.client[0].log.info('Failed to delete sssd cache')
                assert False
            cmd_start = 'systemctl start sssd'
            multihost.client[0].run_command(cmd_start)

        except subprocess.CalledProcessError:
            clear_sssd_cache = False

        if not clear_sssd_cache:
            string = "Numerical result out of range"
            file_content = multihost.client[0].get_file_contents(logfile)
            x = string.encode('utf-8') in file_content
            if x is True:
                assert True
            else:
                assert False
            return

    def test_0004_timeoutminus100(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test of invalid value of ldap connection timeout that is
         -100 in our case.
        :id: d68b6d42-30bd-4abb-bbf4-363388da931d
        :expectedresults: It shoud instatly release the
         connection after establishing
        """
        multihost.client[0].log.info(
            '\n\n\nTesting for ldap_connection_expire_'
            'timeout = -100')
        domain_params = {
            'ldap_uri': 'ldap://%s' % (multihost.master[0].sys_hostname)}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        domain_params = {'ldap_connection_expire_timeout': -100}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)

        self.run_test(-100, multihost)

    def test_0005_timeout0(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test for value of ldap connection timeout 0.
        :id: 39af02aa-0860-4189-afc9-3ead42fd5fc1
        :expectedresults: It should have to release
         the connection instantly after establishing
        """
        multihost.client[0].log.info(
            '\n\n\nTesting for ldap_connection_expire_'
            'timeout = 0')
        domain_params = {
            'ldap_uri': 'ldap://%s' % (multihost.master[0].sys_hostname)}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)
        domain_params = {'ldap_connection_expire_timeout': 0}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params)

        self.run_test(0, multihost)

    def run_test(self, timeout, multihost):
        """
        Runs the remaining test
        :param str timeout:takes the vlalue of timeout for ldap
         and string 'krb' in case of kerberos
        :param obj multihost: multihost object

        :Steps:
              1. Setup ldap_connection_expire_timeout to a certain timeout. For
              Kerberos, this is redundant as connection expires as soon as the
              ticked expires.
              2. Lookup a user and get the port number and sleep for the
              timeout period.
              3. Lookup another user and get the port number.
              4. Compare the 2 port numbers.
        """
        tools = sssdTools(multihost.client[0])
        domainname = tools.get_domain_section_name()
        tools.delete_sssd_domain_log(domainname)
        logfile = '/var/log/sssd/sssd_%s.log' % ds_instance_name

        sssdTools(multihost.client[0]).clear_sssd_cache()

        if timeout == 'krb':
            timeout = 120
        else:
            string = "Option ldap_connection_expire_timeout has value %s" % \
                     timeout
            file_content = multihost.client[0].get_file_contents(logfile)
            x = string.encode('utf-8') in file_content
            if x is True:
                assert True
            else:
                assert False
        lookup_u = 'getent passwd foo1@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(lookup_u)
        assert cmd.returncode == 0

        def find_local_port():
            nsreport = multihost.client[0].run_command(
                ["ss", "-ant"], log_stdout=False).stdout_text
            lines = nsreport.splitlines()
            lines1 = []

            for i in lines:
                if i.find('389') != -1 and i.find('ESTAB') != -1:
                    lines1.append(i)
            del lines

            if len(lines1) > 1:
                assert False

            lines1 = lines1[0]
            port = lines1[
                lines1.find(':') + 1: lines1.find(' ', lines1.find(':'))
            ]
            return int(port)

        localport1 = find_local_port()

        time.sleep(timeout + 5) if timeout > 0 else time.sleep(5)

        lookup_u = 'getent passwd foo2@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(lookup_u)
        assert cmd.returncode == 0

        localport2 = find_local_port()

        assert localport1 != localport2
        if timeout > 0:
            string = "Connection is about to expire, releasing it"
            file_content = multihost.client[0].get_file_contents(logfile)
            x = string.encode('utf-8') in file_content
            if x is True:
                assert True
            else:
                assert False
        cmd_remove_log = "rm /var/log/sssd/sssd_example1.log"
        multihost.client[0].run_command(cmd_remove_log)



""" Automation of Auto Mount suite
:requirement: Ldap Provider - automount
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import time
import pytest
import os
import ldap
import subprocess
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_instance_name, ds_suffix, ds_rootdn, ds_rootpw
from sssd.testlib.common.paths import SSSD_DEFAULT_CONF


def find_logs(multihost, log_name, string_name):
    """This function will find strings in a log file
    log_name: Absolute path of log where the search will happen.
    string_name: String to search in the log file.
    """
    log_str = multihost.client[0].get_file_contents(log_name).decode('utf-8')
    assert string_name in log_str


def clear_only_domain_log(multihost):
    """
    This function will clear domain logs
    """
    client = multihost.client[0]
    log_ssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
    client.run_command(f'> {log_ssd}')


@pytest.fixture(scope='function')
def common_sssd_setup(multihost):
    """
    This is common sssd setup used in this test suite.
    """
    tools = sssdTools(multihost.client[0])
    tools.sssd_conf("nss", {'filter_groups': 'root',
                            'filter_users': 'root',
                            'debug_level': '9'}, action='update')
    tools.sssd_conf("sssd", {'services': 'nss, pam, autofs'}, action='update')
    tools.sssd_conf("autofs", {}, action='update')
    ldap_params = {'enumerate': True, 'use_fully_qualified_names': False}
    tools.sssd_conf(f'domain/{ds_instance_name}', ldap_params)
    tools.clear_sssd_cache()


@pytest.fixture(scope='function')
def ldap_autofs(multihost, request):
    """
    This is common sssd setup used in this test suite.
    """
    tools = sssdTools(multihost.client[0])
    multihost.client[0].run_command("mount")
    tools.sssd_conf("nss", {'filter_groups': 'root',
                            'filter_users': 'root',
                            'debug_level': '9'}, action='update')
    tools.sssd_conf("sssd", {'services': 'nss, pam, autofs'}, action='update')
    tools.sssd_conf("autofs", {'debug_level': "9"}, action='update')
    tools.sssd_conf("domain/example1", {'enumerate': True,
                                        'autofs_provider': "ldap",
                                        'ldap_autofs_search_base': f'ou=mount,{ds_suffix}',
                                        'use_fully_qualified_names': False,
                                        'ldap_autofs_map_object_class': "nisMap",
                                        "ldap_autofs_map_name": "nisMapName",
                                        "ldap_autofs_entry_object_class": "nisObject",
                                        "ldap_autofs_entry_key": "cn",
                                        "ldap_autofs_entry_value": "nisMapEntry"}, action='update')
    tools.clear_sssd_cache()

    def restore():
        """Will restore client after test."""
        multihost.client[0].run_command("systemctl restart autofs", raiseonerr=False)
    request.addfinalizer(restore)

@pytest.fixture(scope='class')
def nfs_server_setup(multihost, request):
    """
    This function will setup NFS server in master machine.
    """
    master = multihost.master[0]
    master.run_command("mkdir -p /export/project{1..2}")
    master.run_command("mkdir /export/directtest{1..2}")
    master.run_command("mkdir /export/projects")
    master.run_command("mkdir /export/projects_new")
    master.run_command("mkdir /export/largepathfolder")
    master.run_command("mkdir -p /export/shared{1..2}/key{1..2}")
    master.run_command("mkdir -p /export/home/test")
    master.run_command('echo "/export    *(rw,sync)" > /etc/exports')
    master.run_command("systemctl restart nfs-server")

    def restore():
        """Will restore nfs server after test."""
        master.run_command("rm -rvf /export/*")
    request.addfinalizer(restore)


@pytest.fixture(scope='class')
def create_users(multihost, request):
    """
    This function will create users for this test.
    """
    client = multihost.client[0]
    client.run_command("authselect select sssd --force")
    client.run_command("cp -f /etc/nsswitch.conf /etc/nsswitch.conf.backup")
    client.run_command("cp -f /etc/sysconfig/autofs /etc/sysconfig/autofs_bkp")
    client.run_command("sed --follow-symlinks -i 's/automount:  files/automount:  sss files/g' /etc/nsswitch.conf")
    ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    ldap_inst.org_unit("mount", ds_suffix)
    user_dn = f'nisMapName=auto.master,ou=mount,{ds_suffix}'
    user_info = {'nisMapName': 'auto.master'.encode('utf-8'),
                 'objectClass': ['nisMap'.encode('utf-8'),
                                 'top'.encode('utf-8')]}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/-,nisMapName=auto.master,ou=mount,{ds_suffix}'
    user_info = {'cn': '/-'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': 'auto.direct'.encode('utf-8'),
                 'nisMapName': 'auto.master'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/share1,nisMapName=auto.master,ou=mount,{ds_suffix}'
    user_info = {'cn': '/share1'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': 'auto.share1'.encode('utf-8'),
                 'nisMapName': 'auto.master'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/share2,nisMapName=auto.master,ou=mount,{ds_suffix}'
    user_info = {'cn': '/share2'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': 'auto.share2'.encode('utf-8'),
                 'nisMapName': 'auto.master'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/export/home,nisMapName=auto.master,ou=mount,{ds_suffix}'
    user_info = {'cn': '/export/home'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': 'auto.home'.encode('utf-8'),
                 'nisMapName': 'auto.master'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'nisMapName=auto.direct,ou=mount,{ds_suffix}'
    user_info = {'objectClass': ['nisMap'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapName': 'auto.direct'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'nisMapName=auto.share1,ou=mount,{ds_suffix}'
    user_info = {'objectClass': ['nisMap'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapName': 'auto.share1'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'nisMapName=auto.share2,ou=mount,{ds_suffix}'
    user_info = {'objectClass': ['nisMap'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapName': 'auto.share2'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    master_server = multihost.master[0].sys_hostname
    user_dn = f'cn=/proj1,nisMapName=auto.direct,ou=mount,{ds_suffix}'
    user_info = {'cn': '/proj1'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/project1'.encode('utf-8'),
                 'nisMapName': 'auto.direct'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/proj2,nisMapName=auto.direct,ou=mount,{ds_suffix}'
    user_info = {'cn': '/proj2'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/project2'.encode('utf-8'),
                 'nisMapName': 'auto.direct'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}'
    user_info = {'cn': '/folder1/folder2/projects'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/projects'.encode('utf-8'),
                 'nisMapName': 'auto.direct'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/directtest,nisMapName=auto.direct,ou=mount,{ds_suffix}'
    user_info = {'cn': '/directtest'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/directtest1'.encode('utf-8'),
                 'nisMapName': 'auto.direct'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=key1,nisMapName=auto.share1,ou=mount,{ds_suffix}'
    user_info = {'cn': 'key1'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/shared1/key1'.encode('utf-8'),
                 'nisMapName': 'auto.share1'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=key2,nisMapName=auto.share1,ou=mount,{ds_suffix}'
    user_info = {'cn': 'key2'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/shared1/key2'.encode('utf-8'),
                 'nisMapName': 'auto.share1'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=key1,nisMapName=auto.share2,ou=mount,{ds_suffix}'
    user_info = {'cn': 'key1'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/shared2/key1'.encode('utf-8'),
                 'nisMapName': 'auto.share2'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'nisMapName=auto.home,ou=mount,{ds_suffix}'
    user_info = {'nisMapName': 'auto.home'.encode('utf-8'),
                 'objectClass': ['nisMap'.encode('utf-8'),
                                 'top'.encode('utf-8')]}
    ldap_inst.add_entry(user_info, user_dn)
    user_dn = f'cn=/,nisMapName=auto.home,ou=mount,{ds_suffix}'
    user_info = {'cn': '/'.encode('utf-8'),
                 'objectClass': ['nisObject'.encode('utf-8'),
                                 'top'.encode('utf-8')],
                 'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/home/&'.encode('utf-8'),
                 'nisMapName': 'auto.home'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    file_location_cpp = '/script/a1.cpp'
    client.transport.put_file(os.path.dirname(os.path.abspath(__file__)) + file_location_cpp, '/tmp/a1.cpp')

    def restore():
        """
        Delete user after test finish
        """
        client.run_command("cp -f /etc/nsswitch.conf.backup /etc/nsswitch.conf")
        client.run_command("cp -vf /etc/sysconfig/autofs_bkp /etc/sysconfig/autofs")
        client.run_command("rm -vf /tmp/a1.cpp")
        for i in range(250):
            user_dn = f'cn=/foo{i},nisMapName=auto.master,ou=mount,{ds_suffix}'
            ldap_inst.del_dn(user_dn)
            user_dn = f'cn=testkey{i},nisMapName=auto.share,ou=mount,{ds_suffix}'
            ldap_inst.del_dn(user_dn)
        for dn_dn in [f'nisMapName=auto.share,ou=mount,{ds_suffix}',
                      f'cn=/largepathfolder1/largepathfolder2/largepathfolder3/'
                      f'largepathfolder4/largepathfolder5/largepathfolder6/largepathfolder7/'
                      f'largepathfolder8/largepathfolder9/largepathfolder10/largepathfolder11/'
                      f'largepathfolder12/largepathfolder13/largepathfolders,'
                      f'nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'cn=key1,nisMapName=auto.share3,ou=mount,{ds_suffix}',
                      f'cn=/share3,nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.share3,ou=mount,{ds_suffix}']:
            ldap_inst.del_dn(dn_dn)
        time.sleep(2)
        for dn_dn in [f'cn=/,nisMapName=auto.home,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.home,ou=mount,{ds_suffix}',
                      f'cn=key1,nisMapName=auto.share2,ou=mount,{ds_suffix}',
                      f'cn=key2,nisMapName=auto.share1,ou=mount,{ds_suffix}',
                      f'cn=key1,nisMapName=auto.share1,ou=mount,{ds_suffix}',
                      f'cn=/directtest,nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'cn=/proj2,nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'cn=/proj1,nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.share2,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.share1,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.direct,ou=mount,{ds_suffix}',
                      f'cn=/export/home,nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'cn=/share2,nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'cn=/share1,nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'cn=/-,nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'nisMapName=auto.master,ou=mount,{ds_suffix}',
                      f'ou=mount,{ds_suffix}']:
            ldap_inst.del_dn(dn_dn)
    request.addfinalizer(restore)


@pytest.mark.tier1_4
@pytest.mark.usefixtures('setup_sssd',
                         'create_users',
                         'nfs_server_setup')
@pytest.mark.offline
class TestAutoFs(object):
    """
    This is test case class for auto fs suite
    """
    @staticmethod
    def test_autofs_search(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: ldap autofs search base not mentioned
        :id: 838bd768-9897-11ed-a1c6-845cf3eff344
        :setup:
            1. SSSD configuration is updated to delete the ldap_search_base option.
            2. The SSSD cache is cleared.
        :steps:
            1. Sssd default autofs config
            2. Dont configure autofs search base with sssd
        :expectedresults:
            1. Sssd should work without autofs search base
            2. Corresponding sssd logs should be generated
        """
        tools = sssdTools(multihost.client[0])
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        tools.sssd_conf("domain/example1", {'ldap_search_base': ''}, action='delete')
        tools.clear_sssd_cache()
        find_logs(multihost, log_sssd, "Got rootdse")
        find_logs(multihost, log_sssd, f"Setting option [ldap_autofs_search_base] to [{ds_suffix}]")

    @staticmethod
    def test_ldap_search_base(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: ldap autofs search base picks up the value from ldap search base
        :id: 7fc8e7f6-9897-11ed-8ec5-845cf3eff344
        :setup:
            1. SSSD configuration is updated to ldap_search_base.
            2. The SSSD cache is cleared.
        :steps:
            1. Sssd should work with autofs search base
        :expectedresults:
            1. Corresponding sssd logs should be generated
        """
        tools = sssdTools(multihost.client[0])
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        ldap_params = {'ldap_search_base': f'ou=mount,{ds_suffix}'}
        tools.sssd_conf(f'domain/{ds_instance_name}', ldap_params)
        tools.clear_sssd_cache()
        find_logs(multihost, log_sssd, f"Option ldap_autofs_search_base set to ou=mount,{ds_suffix}")

    @staticmethod
    def test_ldap_autofs_search_base(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: ldap autofs search base set to ou equals to mount and DS BASE DN
        :id: 7bbdddb0-9897-11ed-85fc-845cf3eff344
        :setup:
            1. SSSD configuration is updated for ldap_search_base to ldap_autofs_search_base
            2. The SSSD cache is cleared.
        :steps:
            1. Remove ldap_search_base
        :expectedresults:
            1. Sssd should work with autofs ldap_autofs_search_base and
                Corresponding sssd logs should be generated
        """
        tools = sssdTools(multihost.client[0])
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        # ldap autofs search base set to ou equals to mount and DS BASE DN
        client = multihost.client[0]
        tools.sssd_conf("domain/example1", {'ldap_search_base': f'ou=mount,{ds_suffix}'}, action='update')
        client.run_command("sed -i 's/ldap_search_base/ldap_autofs_search_base/g' /etc/sssd/sssd.conf")
        tools.clear_sssd_cache()
        find_logs(multihost, log_sssd, f"Search base added: [AUTOFS][ou=mount,{ds_suffix}][SUBTREE]")

    @staticmethod
    def test_autofs_provider_none(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: ldap autofs search base set to ou equals to mount and DS
            BASE DN and autofs_provider is None
        :id: 778f7f78-9897-11ed-ba16-845cf3eff344
        :setup:
            1. SSSD configuration is updated for ldap_autofs_search_base
            2. The SSSD cache is cleared.
        :steps:
            1. Set autofs provider equals to none
        :expectedresults:
            1. Sssd should not work with provider none
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'ldap_autofs_search_base':
                                                f'ou=mount,{ds_suffix}'}, action='update')
        ldap_params = {'autofs_provider': None}
        tools.sssd_conf(f'domain/{ds_instance_name}', ldap_params)
        tools.clear_sssd_cache()
        # Automount should not work here
        assert "share1" not in client.run_command("automount -m").stdout_text

    @staticmethod
    def test_autofs_id_provider(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: autofs provider uses the value from id provider
        :id: 738825ec-9897-11ed-ac02-845cf3eff344
        :setup:
            1. SSSD configuration is updated for ldap_autofs_search_base and autofs_provider
            2. The SSSD cache is cleared.
        :steps:
            1. Remove autofs_provider = None from sssd config
        :expectedresults:
            1. Sssd default autofs config works and Corresponding sssd logs should be generated
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'ldap_autofs_search_base': f'ou=mount,{ds_suffix}'}, action='update')
        tools.sssd_conf("domain/example1", {'autofs_provider': None}, action='delete')
        tools.clear_sssd_cache()
        client.run_command("automount -m 1> /tmp/automount")
        find_logs(multihost, "/tmp/automount", "Mount point: /share2")

    @staticmethod
    def test_autofs_provider_ldap(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: Set autofs provider to ldap
        :id: 9304c416-9c70-11ed-9d67-845cf3eff344
        :setup:
            1. SSSD configuration is updated for ldap_autofs_search_base and autofs_provider
            2. The SSSD cache is cleared.
        :steps:
            1. Set autofs provider to ldap, check share2 is visible
        :expectedresults:
            1. Sssd should work with autofs provider to ldap, share2 should visible
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {
            'autofs_provider': 'ldap',
            'ldap_autofs_search_base': f'ou=mount,{ds_suffix}'}, action='update')
        tools.clear_sssd_cache()
        assert "share2" in client.run_command("automount -m").stdout_text

    @staticmethod
    def test_available_automount_maps(multihost, backupsssdconf, ldap_autofs):
        """
        :title: Set autofs provider to ldap
        :id: 6f735d8c-9897-11ed-8c51-845cf3eff344
        :steps:
            1. "automount -m" command, which lists the available automount maps.
                It checks that the maps "share2", "share", "export", and "directtest"
                are all present in the output.
            2. Searching the logs of the SSSD service for specific configuration options and
                verifying that these options have the expected values.
                The options being checked include "ldap_autofs_map_object_class",
                "ldap_autofs_map_name", "ldap_autofs_entry_object_class",
                "ldap_autofs_entry_key", and "ldap_autofs_entry_value".

        :expectedresults:
            1. If any of these maps are not present in the output,
                it raises an exception with a descriptive error message.
            2. If the logs do not contain the expected values for these options,
                an error will be raised.
        """
        # Set autofs provider to ldap
        client = multihost.client[0]
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        assert "share2" in client.run_command("automount -m").stdout_text
        # non default values of all ldap attributes bz822404
        find_logs(multihost, log_sssd, "Option ldap_autofs_map_object_class has value nisMap")
        find_logs(multihost, log_sssd, "Option ldap_autofs_map_name has value nisMapName")
        find_logs(multihost, log_sssd, "Option ldap_autofs_entry_object_class has value nisObject")
        find_logs(multihost, log_sssd, "Option ldap_autofs_entry_key has value cn")
        find_logs(multihost, log_sssd, "Option ldap_autofs_entry_value has value nisMapEntry")
        cmd = client.run_command("automount -m")
        assert "share" in cmd.stdout_text, "Could not find share in automounts"
        assert "export" in cmd.stdout_text, "Could not find export in automounts"
        assert "directtest" in cmd.stdout_text, "Could not find directtest in automounts"

    @staticmethod
    def test_autofs_shares(multihost, backupsssdconf, ldap_autofs):
        """
        :title: Set autofs provider to ldap
        :id: 6b34da0c-9897-11ed-bb11-845cf3eff344
        :steps:
            1. Restart the autofs service on the client machine
            2. Try to access each of the autofs mounts by changing to the respective
                directories and running the cd command twice to return to the current directory.
            3. Check if the mount points are correctly mounted by running the mount
                command and checking the output for specific strings indicating the mount points.
        :expectedresults:
            1. Restart of autofs service should success
            2. Mount Points should accessible
            3. The code checks for the presence of the following mounts:
                /export/home/test
                /export/shared1/key1
                /export/shared1/key2
                /export/shared2/key1
                /export/directtest1
                /export/projects
        """
        # mount autofs shares
        client = multihost.client[0]
        master_server = multihost.master[0].sys_hostname
        client.run_command("service autofs restart")
        # Try to access the autofs mounts
        client.run_command("cd /export/home/test;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/home/test" in cmd.stdout_text, "Could not find mount /export/home/test"
        client.run_command("cd /share1/key1;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/shared1/key1" in cmd.stdout_text, "Could not find mount /export/shared1/key1"
        client.run_command("cd /share1/key2;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/shared1/key2" in cmd.stdout_text, "Could not find mount /export/shared1/key2"
        client.run_command("cd /share2/key1;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/shared2/key1" in cmd.stdout_text, "Could not find mount /export/shared2/key1"
        # Duplicate Direct Mounts
        client.run_command("cd /directtest;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/directtest1" in cmd.stdout_text, "Could not find mount /export/directtest1"
        # Direct Mount
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert f"{master_server}:/export/projects" in cmd.stdout_text, "Could not find mount /export/projects"

    @staticmethod
    def test_autofs_timeout(multihost, backupsssdconf, ldap_autofs):
        """
        :title: Set entry cache autofs timeout value
        :id: 670cf00e-9897-11ed-8169-845cf3eff344
        :Setup:
            1. Update the 'entry_cache_autofs_timeout' value in the SSSD
                configuration for the "domain/example1" domain to 100.
            2. Clears the SSSD cache.
            3. Update the autofs configuration to set the
                timeout to 10 seconds and disables browsing
                MOUNT_NFS_DEFAULT_PROTOCOL to 4
            4. Restarts the autofs service.
        :steps:
            1. Tries to access the "/share2/key1" directory and checks if it
                is mounted from the master server at "/export/shared2/key1".
            2. Searches for log messages indicating that SSSD is searching for automount map entries.
            3. After a delay of 12 seconds, clears the SSSD log file.
            4. Tries to access the "/share2/key1" directory and checks if it is still mounted.
            5. Raises an error if it finds log messages indicating that
                SSSD is searching for automount map entries.
            6. After a delay of 110 seconds, clears the SSSD log file.
            7. Tries to access the "/share2/key1" directory and checks if it is still mounted.
        :expectedresults:
            1. /shared2/key1 should mount
            2. Corresponding sssd logs should be generated
            3. Sssd logs should be cleared
            4. /export/shared2/key1 should be accessible
            5. Sssd should not sent new request for automount map entries
            6. SSSD logs should be cleared
            7. /shared2/key1 should be accessible
        """
        # Set entry cache autofs timeout value
        client = multihost.client[0]
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        tools = sssdTools(multihost.client[0])
        master_server = multihost.master[0].sys_hostname
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "100"}, action='update')
        tools.clear_sssd_cache()
        client.run_command("echo 'TIMEOUT=10' >> /etc/sysconfig/autofs")
        client.run_command("echo 'BROWSE_MODE=\"no\"' >> /etc/sysconfig/autofs")
        client.run_command("echo 'MOUNT_NFS_DEFAULT_PROTOCOL=4' >> /etc/sysconfig/autofs")
        client.run_command("service autofs restart")
        client.run_command("cd /share2/key1;cd -")
        assert f"{master_server}:/export/shared2/key1" in client.run_command("mount").stdout_text
        time.sleep(2)
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        time.sleep(12)
        client.run_command(f'> {log_sssd}')
        client.run_command("cd /share2/key1;cd -")
        assert f"{master_server}:/export/shared2/key1" in client.run_command("mount").stdout_text
        with pytest.raises(AssertionError):
            find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        time.sleep(110)
        client.run_command(f'> {log_sssd}')
        client.run_command("cd /share2/key1;cd -")
        assert f"{master_server}:/export/shared2/key1" in client.run_command("mount").stdout_text

    @staticmethod
    def test_key_location(multihost, backupsssdconf, ldap_autofs):
        """
        :title: change key and location
        :id: 62521da0-9897-11ed-a276-845cf3eff344
        :setup:
            1. Modify the SSSD configuration file.
            2. The cache of the SSSD is then cleared.
            3. The autofs timeout is increased and the autofs service is restarted.
        :steps:
            1. The client system tries to mount a folder (/folder1/folder2/projects)
                from the master server (master_server).
            2. Searches the log files of the SSSD for log entries related to automount map entries.
            3. The mounted folder is then unmounted and the LDAP entry for the folder is updated.
            4. The autofs service is restarted and the client system tries to mount the folder again.
            5. Finally, the test case searches the SSSD log files again to see if the new mount is reflected.
        :expectedresults:
            1. Mount of /folder1/folder2/projects Should Success
            2. Corresponding sssd logs should be generated
            3. Unmount Should Success
            4. Mounting of folder should success
            5. New mount is reflected Successfully
        """
        # change key and location
        client = multihost.client[0]
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        tools = sssdTools(multihost.client[0])
        master_server = multihost.master[0].sys_hostname
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "60"}, action='update')
        tools.clear_sssd_cache()
        client.run_command("sed -i 's/TIMEOUT=10/TIMEOUT=60/g' /etc/sysconfig/autofs")
        client.run_command("service autofs restart")
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert "/folder1/folder2/projects" in cmd.stdout_text
        assert f"{master_server}:/export/projects" in cmd.stdout_text
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        client.run_command("umount -v /folder1/folder2/projects")
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        modify_gid = [(ldap.MOD_REPLACE, 'nisMapEntry',
                       f'-fstype=nfs,rw {master_server}:/export/projects_new'.encode('utf-8'))]
        ldap_inst.modify_ldap(f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}', modify_gid)
        time.sleep(40)
        clear_only_domain_log(multihost)
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert "/folder1/folder2/projects" in cmd.stdout_text
        assert f"{master_server}:/export/projects" in cmd.stdout_text
        with pytest.raises(AssertionError):
            find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        clear_only_domain_log(multihost)
        time.sleep(30)
        client.run_command("umount -v /folder1/folder2/projects")
        # Restart autofs to clear the autofs cache and pull in the changes from ldap
        client.run_command("service autofs restart")
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert "/folder1/folder2/projects" in cmd.stdout_text
        assert f"{master_server}:/export/projects_new" in cmd.stdout_text
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        modify_gid = [(ldap.MOD_REPLACE, 'nisMapEntry',
                       f'-fstype=nfs,rw {master_server}:/export/projects'.encode('utf-8'))]
        ldap_inst.modify_ldap(f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}', modify_gid)

    @staticmethod
    def test_change_key_map(multihost, backupsssdconf, ldap_autofs):
        """
        :title: change key and map
        :id: 5e15fd9c-9897-11ed-acf5-845cf3eff344
        :setup:
            1. Configures SSSD by updating the SSSD configuration file and clearing the SSSD cache.
            3. Restarts the autofs service.
        :steps:
            1. Searches the log files for messages to verify that the mount was successful.
            2. Unmounts the file system and deletes the corresponding automount map from the LDAP server.
            3. Verifies that the unmount was successful and that the NFS file system is no longer accessible.
            4. Restarts the autofs service again and verifies that the NFS file system cannot be accessed.
            5. Adds the deleted automount map back to the LDAP server.
        :expectedresults:
            1. Corresponding sssd logs should be generated
            2. Unmount Should Success
            3. Unmounted NFS file system not be accessible
            4. After restart unmounted NFS file system not be accessible
            5. Deleted mounts can be remounted
        """
        # change key and map
        client = multihost.client[0]
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        master_server = multihost.master[0].sys_hostname
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "60"}, action='update')
        tools.clear_sssd_cache()
        client.run_command("service autofs restart")
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert "/folder1/folder2/projects" in cmd.stdout_text
        assert f"{master_server}:/export/projects" in cmd.stdout_text
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        client.run_command("umount /folder1/folder2/projects")
        ldap_inst.del_dn(f"cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}")
        time.sleep(40)
        clear_only_domain_log(multihost)
        client.run_command("cd /folder1/folder2/projects;cd -")
        cmd = client.run_command("mount")
        assert "/folder1/folder2/projects" in cmd.stdout_text
        assert f"{master_server}:/export/projects" in cmd.stdout_text
        with pytest.raises(AssertionError):
            find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        clear_only_domain_log(multihost)
        time.sleep(30)
        client.run_command("umount -v /folder1/folder2/projects")
        # Restart autofs to clear the autofs cache and pull in the changes from ldap
        client.run_command("service autofs restart")
        time.sleep(10)
        with pytest.raises(subprocess.CalledProcessError):
            client.run_command("ls -d /folder1/folder2/projects")
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        # Adding the deleted direct map
        user_dn = f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}'
        user_info = {'cn': '/folder1/folder2/projects'.encode('utf-8'),
                     'objectClass': ['nisObject'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapEntry':f'-fstype=nfs,rw {master_server}:/export/projects'.encode('utf-8'),
                     'nisMapName': 'auto.direct'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)

    @staticmethod
    def test_bz870045(multihost, backupsssdconf, ldap_autofs):
        """
        :title: Always reread master map from ldap bz870045
        :id: 5983c6f6-9897-11ed-9101-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=870045
        :setup:
            1. Configures SSSD by updating the SSSD configuration file and clearing the SSSD cache.
        :steps:
            1. The client restarts the autofs service
            2. Logs are checked for various messages to confirm the correct behavior of SSSD and automount.
            3. Adds a new map auto.share3 to the LDAP server
            4. Restarts the autofs service on the client
            5. Checks the logs again to confirm the correct behavior of SSSD and automount with the new map.
        :expectedresults:
            1. Restarts should Success
            2. Corresponding logs should generate
            3. auto.share3 should added to master server
            4. Restart should success
            5. Corresponding logs should generate
        """
        # Always reread master map from ldap bz870045
        client = multihost.client[0]
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        master_server = multihost.master[0].sys_hostname
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "60"}, action='update')
        tools.clear_sssd_cache()
        client.run_command("automount -m")
        time.sleep(2)
        find_logs(multihost, "/var/log/sssd/sssd_autofs.log", "Obtaining autofs map auto.master")
        find_logs(multihost, log_sssd, "Requested refresh for: auto.master" )
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        find_logs(multihost, log_sssd, f"{master_server}:/export/shared2/key1")
        time.sleep(30)
        clear_only_domain_log(multihost)
        client.run_command("> /var/log/sssd/sssd_autofs.log")
        client.run_command("service autofs restart")
        time.sleep(2)
        find_logs(multihost, "/var/log/sssd/sssd_autofs.log", "Looking up [auto.master] in data provider")
        find_logs(multihost, "/var/log/sssd/sssd_autofs.log", "Obtaining autofs map auto.master")
        find_logs(multihost, log_sssd, "Requested refresh for: auto.master")
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        time.sleep(10)
        clear_only_domain_log(multihost)
        client.run_command("> /var/log/sssd/sssd_autofs.log")
        ## Add a new map auto.share3
        user_dn = f'cn=/share3,nisMapName=auto.master,ou=mount,{ds_suffix}'
        user_info = {'cn': '/share3'.encode('utf-8'),
                     'objectClass': ['nisObject'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapEntry':'auto.share3'.encode('utf-8'),
                     'nisMapName': 'auto.master'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        user_dn = f'nisMapName=auto.share3,ou=mount,{ds_suffix}'
        user_info = {'objectClass': ['nisMap'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapName': 'auto.share3'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        user_dn = f'cn=key1,nisMapName=auto.share3,ou=mount,{ds_suffix}'
        user_info = {'cn':'key1'.encode('utf-8'),
                     'objectClass': ['nisObject'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapEntry':f'-fstype=nfs,rw {master_server}:/export/shared3/key1'.encode('utf-8'),
                     'nisMapName': 'auto.share3'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        client.run_command("service autofs restart")
        time.sleep(2)
        find_logs(multihost, "/var/log/sssd/sssd_autofs.log", "Obtaining autofs map auto.master")
        find_logs(multihost, log_sssd, "Requested refresh for: auto.master")
        find_logs(multihost, log_sssd, f"Searching for automount map entries with base [ou=mount,{ds_suffix}]")
        find_logs(multihost, log_sssd, f"[cn=/share3,nisMapName=auto.master,ou=mount,{ds_suffix}]")
        find_logs(multihost, log_sssd, f"Adding autofs entry [/share3] - [auto.share3]")

    @staticmethod
    def test_bz876531_bz894428(multihost, backupsssdconf, ldap_autofs):
        """
        :title: sss cache does not work for automount maps bz876531 bz894428
        :id: 53c18da2-9897-11ed-b023-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=894428
                   https://bugzilla.redhat.com/show_bug.cgi?id=876531
        :setup:
            1. Updates the SSSD configuration file with a new setting for entry_cache_autofs_timeout,
                which controls the cache timeout for autofs maps in SSSD.
            2. clear the SSSD cache
        :steps:
            1. Run a series of commands to check that the autofs maps are correctly populated in the SSSD
                cache and the cache entries are not expired.
                search the specified LDB (Lightweight Directory-Based Authentication) database
                (/var/lib/sss/db/cache_example1.ldb) for the specified entries
                (name=auto.master and name=auto.direct) and retrieve their dataExpireTimestamp values.
            2. Search for some errors which should not occur.
            3. Modify a direct map
            4. on the client run sss_cache -A
            5. on the client run automount -m to see if the change is visible by the automounter
        :expectedresults:
            1. Retrieving dataExpireTimestamp values should success
            2. There should not be any error
            3. Direct map modification should success
            4. sss_cache -A command should success
            5. automount -m should display the current automount maps
        """
        # sss cache does not work for automount maps bz876531 bz894428
        client = multihost.client[0]
        master_server = multihost.master[0].sys_hostname
        log_sssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "60"}, action='update')
        tools.clear_sssd_cache()
        client.run_command("service autofs restart")
        time.sleep(2)
        client.run_command("automount -m")
        for mount_point in ["auto.master",
                            "auto.direct",
                            "auto.home",
                            "auto.share1",
                            "auto.share2",
                            "auto.share3"]:
            assert "dataExpireTimestamp" in \
                   client.run_command(f"ldbsearch -H "
                                      f"/var/lib/sss/db/cache_example1.ldb name={mount_point} "
                                      f"dataExpireTimestamp").stdout_text
        client.run_command("sss_cache -A --domain=example1 --debug 10 2> /tmp/cache_debugout")
        for result in ["failed",
                       "No cache object matched the specified search",
                       "No such autofs map"]:
            with pytest.raises(AssertionError):
                find_logs(multihost, "/tmp/cache_debugout", result)
        for mount_point in ["auto.master",
                            "auto.direct",
                            "auto.home",
                            "auto.share1",
                            "auto.share2",
                            "auto.share3"]:
            assert "dataExpireTimestamp" in \
                   client.run_command(f"ldbsearch -H /var/lib/sss/db/cache_example1.ldb "
                                      f"name={mount_point} dataExpireTimestamp").stdout_text
        client.run_command("automount -m")
        # Modify a direct map
        modify_gid = [(ldap.MOD_REPLACE, 'nisMapEntry',
                       f'-fstype=nfs,rw {master_server}:/export/projects_old'.encode('utf-8'))]
        ldap_inst.modify_ldap(f'cn=/folder1/folder2/projects,nisMapName=auto.direct,ou=mount,{ds_suffix}', modify_gid)
        clear_only_domain_log(multihost)
        client.run_command("sss_cache -a auto.direct --debug 10 2> /tmp/cache_debugout")
        for result in ["failed", "No cache object matched the specified search", "No such autofs map"]:
            with pytest.raises(AssertionError):
                find_logs(multihost, "/tmp/cache_debugout", result)
        for mount_point in ["auto.direct",
                            "auto.master",
                            "auto.home",
                            "auto.share1",
                            "auto.share2",
                            "auto.share3"]:
            assert "dataExpireTimestamp" in \
                   client.run_command(f"ldbsearch -H /var/lib/sss/db/cache_example1.ldb "
                                      f"name={mount_point} dataExpireTimestamp").stdout_text
        client.run_command("automount -m")
        time.sleep(2)
        find_logs(multihost, log_sssd, f"{master_server}:/export/projects_old")
        modify_gid = [(ldap.MOD_REPLACE, 'nisMapEntry',
                       f'-fstype=nfs,rw {master_server}:/export/shared1/key_new'.encode('utf-8'))]
        ldap_inst.modify_ldap(f'cn=key1,nisMapName=auto.share1,ou=mount,{ds_suffix}', modify_gid)
        clear_only_domain_log(multihost)
        client.run_command("sss_cache -a auto.share1 --domain=example1 --debug 10 2> /tmp/cache_debugout")
        for result in ["failed", "No cache object matched the specified search", "No such autofs map"]:
            with pytest.raises(AssertionError):
                find_logs(multihost, "/tmp/cache_debugout", result)
        for mount_point in ["auto.direct",
                            "auto.master",
                            "auto.home",
                            "auto.share1",
                            "auto.share2",
                            "auto.share3"]:
            assert "dataExpireTimestamp" in \
                   client.run_command(f"ldbsearch -H /var/lib/sss/db/cache_example1.ldb "
                                      f"name={mount_point} dataExpireTimestamp").stdout_text
        client.run_command("automount -m")
        time.sleep(5)
        find_logs(multihost, log_sssd, f"{master_server}:/export/shared1/key_new")

    @staticmethod
    def test_bz811987(multihost, backupsssdconf, ldap_autofs):
        """
        :title: maximum key name must be PATH MAX bz811987
        :id: 4e9e6930-9897-11ed-b12f-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=811987
        :setup:
            1. Updates the SSSD configuration file with a new setting for entry_cache_autofs_timeout,
                which controls the cache timeout for autofs maps in SSSD.
            2. clear the SSSD cache
        :steps:
            1. Create a big size user for automount
            2. Mount it and should work as normal.
        :expectedresults:
            1. size user for automount should be added
            2. Should be mounted
        """
        # maximum key name must be PATH MAX bz811987
        client = multihost.client[0]
        master_server = multihost.master[0].sys_hostname
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("domain/example1", {'entry_cache_autofs_timeout': "60"}, action='update')
        tools.clear_sssd_cache()
        user_dn = f'cn=/largepathfolder1/largepathfolder2/largepathfolder3/largepathfolder4/' \
                  f'largepathfolder5/largepathfolder6/largepathfolder7/largepathfolder8/largepathfolder9/' \
                  f'largepathfolder10/largepathfolder11/largepathfolder12/largepathfolder13/largepathfolders,' \
                  f'nisMapName=auto.direct,ou=mount,{ds_suffix}'
        user_info = {'objectClass': ['nisObject'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapName': 'auto.direct'.encode('utf-8'),
                     'cn': '/largepathfolder1/largepathfolder2/largepathfolder3/'
                           'largepathfolder4/largepathfolder5/largepathfolder6/largepathfolder7/'
                           'largepathfolder8/largepathfolder9/largepathfolder10/largepathfolder11/'
                           'largepathfolder12/largepathfolder13/largepathfolders'.encode('utf-8'),
                     'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/largepathfolder'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        client.run_command("service autofs restart")
        tools.clear_sssd_cache()
        client.run_command("cd /largepathfolder1/largepathfolder2/largepathfolder3/largepathfolder4/"
                           "largepathfolder5/largepathfolder6/largepathfolder7/largepathfolder8/"
                           "largepathfolder9/largepathfolder10/largepathfolder11/largepathfolder12/"
                           "largepathfolder13/largepathfolders;cd -")
        assert f"{master_server}:/export/largepathfolder" in client.run_command("mount").stdout_text

    @staticmethod
    def test_autofs_dumpmaps(multihost, backupsssdconf, ldap_autofs):
        """
        :title: SSSD frequently fails to return automount maps from LDAP bz967636
        :id: 499275c6-9897-11ed-8e99-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=967636
        :steps:
            1. Have lots of automount entries in an LDAP server (~250 in our case).
            2. On an SSSD+autofs+LDAP client, run `automount --dumpmaps` 10 times.
            3. assert --dumpmaps is returining complete maps from ldap
        :expectedresults:
            1. 250 automount entries should be created
            2. Sssd should not crash
            3. automount --dumpmaps should always return a
                complete set of automount maps from  LDAP.
        """
        # SSSD frequently fails to return automount maps from LDAP bz967636
        client = multihost.client[0]
        master_server = multihost.master[0].sys_hostname
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        # sssd automount maps from LDAP test BZ 967636
        client.run_command("service autofs restart")
        user_dn = f'nisMapName=auto.share,ou=mount,{ds_suffix}'
        user_info = {'objectClass': ['nisMap'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisMapName': 'auto.share'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        # Test autofs dumpmaps consistency.
        for i in range(250):
            user_dn = f'cn=/foo{i},nisMapName=auto.master,ou=mount,{ds_suffix}'
            user_info = {'objectClass': ['nisObject'.encode('utf-8'),
                                         'top'.encode('utf-8')],
                         'cn': f'/foo{i}'.encode('utf-8'),
                         'nisMapName': 'auto.master'.encode('utf-8'),
                         'nisMapEntry': 'auto.share'.encode('utf-8')}
            ldap_inst.add_entry(user_info, user_dn)
            user_dn = f'cn=testkey{i},nisMapName=auto.share,ou=mount,{ds_suffix}'
            user_info = {'objectClass': ['nisObject'.encode('utf-8'),
                                         'top'.encode('utf-8')],
                         'cn': f'testkey{i}'.encode('utf-8'),
                         'nisMapName': 'auto.share'.encode('utf-8'),
                         'nisMapEntry': f'-fstype=nfs,rw {master_server}:/export/foo/testkey{i}'.encode('utf-8')}
            ldap_inst.add_entry(user_info, user_dn)
        dumpmap_fixed_val = "automount - -dumpmaps | wc - l"
        for i in range(1, 11):
            dumpmap_current_val = "automount - -dumpmaps | wc - l"
            assert dumpmap_fixed_val == dumpmap_current_val

    @staticmethod
    def test_autofs_segfaults(multihost, backupsssdconf):
        """
        :title: Automount segfaults in sss nss check header bz1123291
        :id: 44af3d14-9897-11ed-a25c-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1123291
        :setup:
            1. The original SSSD configuration file is backed up and the new
                configuration options are set using the sssd_conf method.
                The options include the filter groups, filter users, and debug level for the nss section.
                The script then sets additional parameters for the domain specified by ds_instance_name.
            2. The SSSD cache is then cleared using the clear_sssd_cache method, and the autofs service is restarted.
        :steps:
            1. The script compiles and runs a C++ program called a1 100 times and not able to see the crash.
        :expectedresults:
            1. Should not crash
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        client.run_command(f'cp -f {SSSD_DEFAULT_CONF}.orig {SSSD_DEFAULT_CONF}')
        tools.sssd_conf("nss", {'filter_groups': 'root',
                                'filter_users': 'root',
                                'debug_level': '9'}, action='update')
        ldap_params = {'use_fully_qualified_names': False,
                       'ldap_search_base': f'{ds_suffix}',
                       'ldap_autofs_map_object_class': 'nisMap',
                       'ldap_autofs_map_name': 'nisMapName',
                       'ldap_autofs_entry_object_class': 'nisObject',
                       'ldap_autofs_entry_key': 'cn',
                       'ldap_autofs_entry_value': 'nisMapEntry'}
        tools.sssd_conf(f'domain/{ds_instance_name}', ldap_params)
        tools.clear_sssd_cache()
        time.sleep(2)
        client.run_command("service autofs restart")
        client.run_command("g++ -lpthread -o a1 /tmp/a1.cpp")
        client.run_command("chmod +x a1")
        for _ in range(100):
            cmd = client.run_command("./a1")
            assert cmd.returncode == 0



""" Automation of rfc2307

:requirement: rfc2307
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import pytest
from sssd.testlib.common.libkrb5 import krb5srv
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_suffix, ds_rootdn, ds_rootpw
from sssd.testlib.common.exceptions import LdapException
from sssd.testlib.common.ssh2_python import run_command_client


def usr_grp(multihost, obj_info, type):
    """
    Add an object, user or group, in the ldap-server
        :param dict obj_info: an object(user/group) details
        :param str type: Either 'user' or 'group'
        :return: None
        :exception: LdapException
    """
    ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    krb = krb5srv(multihost.master[0], 'EXAMPLE.TEST')
    if type == 'user':
        usr = obj_info.get('uid')
        try:
            if ldap_inst.posix_user("ou=People", ds_suffix, obj_info):
                krb.add_principal(usr, 'user', 'Secret123')
        except LdapException:
            print(f"Unable to add ldap User {obj_info}")
    if type == 'group':
        try:
            ldap_inst.posix_group("ou=Groups", ds_suffix, obj_info,
                                  memberUid=obj_info.get('memberUid'))
        except LdapException:
            print(f"Unable to add ldap group {obj_info}")


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.rfc2307
class Testrfc2307(object):
    """
    This is test case class for ldap rfc2307

        :setup:
          1. Configure SSSD to authenticate against directory server
          2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
    """
    @pytest.mark.converted('test_ldap.py', 'test_ldap__user_with_whitespace')
    @pytest.mark.tier2
    def test_0001_bz1362023(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: rfc2307: user with spaces at beginning
        :id: 6923436c-d4e4-4a0d-a8f3-1e94ecb1dee3
        :description: user with a white space at the beginning in it's name
         should be able to log in
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=1067476
          https://bugzilla.redhat.com/show_bug.cgi?id=1065534
        :steps:
          1. Create user with a white space at beginning in their name
          2. Restart SSSD with cleared cache
          3. Fetch user information using 'id'
          4. Confirm user is able to log in via ssh
          5. A normal user information is fetched
          6. Confirm a user information is not fetched if a space is added
             as it's first character
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          """
        usr = ' tuser'
        usr_info = {'cn': usr, 'uid': usr,
                    'uidNumber': '34583100',
                    'gidNumber': '34564100'}
        usr_grp(multihost, usr_info, 'user')
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        tools.clear_sssd_cache()
        user = f' tuser@{domain_name}'
        id_cmd = run_command_client(multihost, user, "Secret123", 'id')
        assert 'uid=34583100' in id_cmd
        user = f'tuser@{domain_name}'
        cmd = multihost.client[0].run_command(f'id {user}', raiseonerr=False)
        assert cmd.returncode != 0
        user = f'foo1@{domain_name}'
        cmd = multihost.client[0].run_command(f'id {user}', raiseonerr=False)
        assert cmd.returncode == 0
        user = f'\\ foo1@{domain_name}'
        cmd = multihost.client[0].run_command(f'id {user}', raiseonerr=False)
        assert cmd.returncode != 0



""" Automation of misc bugs

:requirement: IDM-SSSD-REQ : LDAP Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

from __future__ import print_function
import re
import time
import subprocess
import pytest
import textwrap
from datetime import datetime as D_T
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_instance_name, ds_suffix, ds_rootdn, ds_rootpw
from sssd.testlib.common.ssh2_python import check_login_client
from sssd.testlib.common.helper_functions import count_pattern_logs, client_remove_file


def find_logs(multihost, log_name, string_name):
    """This function will find strings in a log file
    log_name: Absolute path of log where the search will happen.
    string_name: String to search in the log file.
    """
    log_str = multihost.client[0].get_file_contents(log_name).decode('utf-8')
    return string_name in log_str


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.misc
class TestMisc(object):
    """
    This is for misc bugs automation
    """
    @pytest.mark.tier1
    def test_0001_ldapcachepurgetimeout(self,
                                        multihost, backupsssdconf):
        """
        :title: ldap_purge_cache_timeout validates most of
         the entries once the cleanup task kicks in.
        :id: 5ed009f4-6462-402c-a1fe-4bef7fb0ef73
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1471808
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        domainname = tools.get_domain_section_name()
        section = "domain/%s" % ds_instance_name
        params = {'enumerate': 'True',
                  'ldap_enumeration_refresh_timeout': '30',
                  'ldap_purge_cache_timeout': '60',
                  'entry_cache_timeout': '20'}
        tools.sssd_conf(section, params)
        multihost.client[0].service_sssd('start')
        try:
            multihost.client[0].run_command('id foo1@%s' % domainname,
                                            raiseonerr=False)
        except subprocess.CalledProcessError:
            pytest.fail("Unable to fetch the user foo1@%s" % domainname)
        for i in range(2):
            time.sleep(60)
            log_file = '/var/log/sssd/sssd_%s.log' % domainname
            log_str = multihost.client[0].get_file_contents(log_file)
            log1 = re.compile(r'Found 0 expired user')
            result = log1.search(log_str.decode())
            if result is not None:
                status = 'PASS'
            else:
                status = 'FAIL'
            log2 = re.compile(r'Found [1-9]* expired user')
            result1 = log2.search(log_str.decode())
            if result1 is None:
                status = 'PASS'
            else:
                status = 'FAIL'
        multihost.client[0].service_sssd('stop')
        tools.sssd_conf(section, params, action='delete')
        multihost.client[0].service_sssd('start')
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_0003_sssd_crashes_after_update(self, multihost,
                                            backupsssdconf):
        """
        :title: misc: sssd crashes after last update to
         sssd-common-1.16.4-37.el7_8.1
        :id: 55cbdb9c-c62e-4604-8c77-9d70dd333a50
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1854317
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        client = sssdTools(multihost.client[0])
        domain_params = {'cache_credentials': 'true',
                         'refresh_expired_interval': '4000'}
        client.sssd_conf(f'domain/{domain_name}', domain_params)
        s_param = {'domains': f'{domain_name}, local'}
        tools.sssd_conf('sssd',s_param, action='update')
        param = {'id_provider': 'proxy',
                 'proxy_lib_name': 'files',
                 'passwd_files': '/etc/passwd',
                 'proxy_pam_target': 'sssd-shadowutils'}
        tools.sssd_conf('domain/local', param)
        multihost.client[0].service_sssd('restart')
        user = f'foo1@{domain_name}'
        check_login_client(multihost, user, "Secret123")
        time.sleep(2)
        for _ in range(3):
            check_login_client(multihost, user, "Secret123")
        time.sleep(2)
        check_login_client(multihost, user, "Secret123")
        time.sleep(2)
        cmd_id = f'id {user}'
        cmd = multihost.client[0].run_command(cmd_id)
        if "no such user" in cmd.stdout_text:
            status = "FAIL"
        else:
            status = "PASS"
        assert status == "PASS"

    @pytest.mark.tier1
    def test_0004_sssd_api_conf(self, multihost, backupsssdconf):
        """
        :title: sssd.api.conf and sssd.api.d
         should belong to python-sssdconfig package
        :id: 4c6bd6a2-d7eb-4c2c-9346-a89b2bdd553e
        :description: Verify by removing sssd-common that
         sssd.api.conf, sssd.api.d is part of python-sssdconfig package
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1800564 (RHEL7.8)
         https://bugzilla.redhat.com/show_bug.cgi?id=1829470 (RHEL8.2)
        """
        # remove sssd-common package
        rpm_remove = 'rpm -e sssd-common --nodeps'
        try:
            multihost.client[0].run_command(rpm_remove)
        except subprocess.CalledProcessError:
            print("Failed to remove sssd-common package")
            status = 'FAIL'
        else:
            python_cmd = "python3 -c 'from SSSDConfig import"\
                         " SSSDConfig; print(SSSDConfig());'"
            cmd = multihost.client[0].run_command(python_cmd, raiseonerr=False)
            if cmd.returncode != 0:
                status = 'FAIL'
            else:
                status = 'PASS'
        # reinstall sssd-common
        install = 'yum -y install sssd-common'
        multihost.client[0].run_command(install, raiseonerr=False)
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_0005_getent_homedirectory(self, multihost,
                                       backupsssdconf):
        """
        :title: misc: fallback_homedir returns '/'
         for empty home directories in passwd file
        :id: 69a6b54e-a8eb-4145-8554-c5e666d82276
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1660693
        """
        multihost.client[0].service_sssd('restart')
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_info = {'cn': 'user_exp4'.encode('utf-8'),
                     'objectClass': [b'top', b'person',
                                     b'inetOrgPerson',
                                     b'organizationalPerson',
                                     b'posixAccount'],
                     'sn': 'user_exp'.encode('utf-8'),
                     'uid': 'user_exp'.encode('utf-8'),
                     'userPassword': 'Secret123'.encode('utf-8'),
                     'homeDirectory': ' '.encode('utf-8'),
                     'uidNumber': '121012'.encode('utf-8'),
                     'gidNumber': '121012'.encode('utf-8'),
                     'loginShell': '/bin/bash'.encode('utf-8')}
        user_dn = 'uid=user_exp4,ou=People,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        cmd_getent = "getent passwd -s sss user_exp4@example1"
        cmd = multihost.client[0].run_command(cmd_getent)
        ldap_inst.del_dn(user_dn)
        assert ":/:" not in cmd.stdout_text

    @pytest.mark.tier1_2
    def test_0006_getent_group(self, multihost,
                               backupsssdconf,
                               delete_groups_users):
        """
        :title: 'getent group ldapgroupname' doesn't
         show any LDAP users or some LDAP users when
         'rfc2307bis' schema is used with SSSD
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1817122
        :id: dc81bb8e-72c0-11eb-9eae-002b677efe14
        :customerscenario: true
        :steps:
            1. Configure SSSD with id_provider = ldap and
               set ldap_schema = rfc2307bis
            2. Add necessary users and groups with uniqueMember.
            3. Check 'getent group ldapgroupname' output.
        :expectedresults:
            1. Should succeed
            2. Should succeed
            3. 'getent group ldapgroupname' should show
               all it's member ldapusers.
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_schema': 'rfc2307bis',
                         'ldap_group_member': 'uniquemember'}
        client.sssd_conf(f'domain/{domain_name}', domain_params)
        multihost.client[0].service_sssd('restart')
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_info = {
            'ou': 'Unit1'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'Unit2'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=Unit2,ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'users'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=users,ou=Unit2,ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'posix_groups'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=posix_groups,ou=Unit2,' \
                  'ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'netgroups'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=netgroups,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'services'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=services,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'ou': 'sudoers'.encode('utf-8'),
            'objectClass': [b'top', b'organizationalUnit']}
        user_dn = 'ou=sudoers,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        for i in range(1, 9):
            user_info = {
                'cn': f'user-{i}'.encode('utf-8'),
                'objectClass': [b'top', b'posixAccount'],
                'uid': f'user-{i}'.encode('utf-8'),
                'uidNumber': f'1111{i}'.encode('utf-8'),
                'gidNumber': f'1111{i}'.encode('utf-8'),
                'homeDirectory': f'/home/user-{i}'.encode('utf-8')}
            user_dn = f'cn=user-{i},ou=users,ou=Unit2,' \
                      f'ou=Unit1,dc=example,dc=test'
            (_, _) = ldap_inst.add_entry(user_info, user_dn)
        for i in range(1, 9):
            user_info = {
                'cn': f'user-{i}'.encode('utf-8'),
                'objectClass': [b'top', b'posixGroup'],
                'gidNumber': f'1111{i}'.encode('utf-8')}
            user_dn = f'cn=user-{i},ou=posix_groups,' \
                      f'ou=Unit2,ou=Unit1,dc=example,dc=test'
            (_, _) = ldap_inst.add_entry(user_info, user_dn)
        user_info = {
            'cn': 'group-1'.encode('utf-8'),
            'objectClass': [b'top', b'posixGroup', b'groupOfUniqueNames'],
            'gidNumber': '20001'.encode('utf-8'),
            'uniqueMember': [
                b'cn=user-1,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-3,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-5,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-7,ou=users,ou=unit2,ou=unit1,dc=example,dc=test']}
        user_dn = 'cn=group-1,ou=posix_groups,ou=Unit2,' \
                  'ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)

        user_info = {
            'cn': 'group-2'.encode('utf-8'),
            'objectClass': [b'top', b'posixGroup', b'groupOfUniqueNames'],
            'gidNumber': '20002'.encode('utf-8'),
            'uniqueMember': [
                b'cn=user-2,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-4,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-6,ou=users,ou=unit2,ou=unit1,dc=example,dc=test',
                b'cn=user-8,ou=users,ou=unit2,ou=unit1,dc=example,dc=test']}
        user_dn = 'cn=group-2,ou=posix_groups,ou=Unit2,' \
                  'ou=Unit1,dc=example,dc=test'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        time.sleep(3)
        cmd = multihost.client[0].run_command("getent group "
                                              "group-2@example1")
        assert "group-2@example1:*:20002:user-2@example1," \
               "user-4@example1,user-6@example1," \
               "user-8@example1" in cmd.stdout_text

    @pytest.mark.tier1_2
    def test_0007_getent_admproxy(self, multihost, backupsssdconf):
        """
        :title: 'getent passwd adm@proxy' doesn't return anything when
         'cache_first = True' option is used with nss.
        :description: Lookup with the fully-qualified name of a user or
         group will fail if the requested object is not already in the
         cache.
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2013294
        :id: 9ff64ee0-255d-46ac-bf0a-b022eaad463e
        :customerscenario: false
        :steps:
            1. Configure SSSD with nss having cache_first = True.
            2. restart SSSD with empty cache.
            3. Check 'getent passwd adm@proxy' output.
        :expectedresults:
            1. Should succeed
            2. Should succeed
            3. 'getent passwd adm@proxy' should return identity lookup
               of user adm.
        """
        getent_admproxy = "getent passwd adm@proxy"
        tools = sssdTools(multihost.client[0])
        section = "sssd"
        section_params = {"domains": "proxy", "services": "nss"}
        tools.sssd_conf(section, section_params, action="update")
        section = "domain/proxy"
        section_params = {"id_provider": "proxy", "proxy_lib_name": "files",
                          'auth_provider': "none"}
        tools.sssd_conf(section, section_params, action="update")
        section = "nss"
        section_params = {"cache_first": "True"}
        tools.sssd_conf(section, section_params, action="update")
        tools.clear_sssd_cache(start=True)
        cache_first_true = multihost.client[0].run_command(getent_admproxy,
                                                           raiseonerr=False)
        assert cache_first_true.returncode == 0, "Bug 2013294/1992973/2013379"
        section = "nss"
        section_params = {"cache_first": "True"}
        tools.sssd_conf(section, section_params, action="delete")
        tools.clear_sssd_cache(start=True)
        cache_first_false = multihost.client[0].run_command(getent_admproxy,
                                                            raiseonerr=False)
        assert cache_first_false.returncode == 0

    @staticmethod
    @pytest.mark.tier1_2
    def test_0008_1636002(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: ldap_provider: socket-activated services start as
         the sssd user and then are unable to read the confdb
        :id: 7a33729a-ab74-4d9e-9d75-e952deaa7bd2
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1636002
        :customerscenario: true
        :steps:
            1. Switch to socket activated services, restart sssd
            2. Check 'getent passwd <user> output.
            3. Run ssh for the user to trigger PAM.
            4. Check log for error messages related to opening
               /var/lib/sss/db/config.ldb
        :expectedresults:
            1. No issue switching and sssd has started.
            2. It should succeed.
            3. /var/log/sssd/sssd_pam.log is present
            4. The error messages are not present.
        :teardown:
            1. Undo socket activation.
            2. Restore sssd.conf
        """
        # pylint: disable=unused-argument
        client = sssdTools(multihost.client[0])
        client.clear_sssd_cache()

        domain_name = client.get_domain_section_name()
        user = f'foo1@{domain_name}'

        # Configure socket activation
        sssd_params = {'services': ''}
        client.sssd_conf('sssd', sssd_params)
        client.clear_sssd_cache()
        enable_cmd = "systemctl enable sssd-nss.socket sssd-pam.socket"
        multihost.client[0].run_command(enable_cmd)
        multihost.client[0].service_sssd('restart')

        # Show the sssd config
        multihost.client[0].run_command(
            'cat /etc/sssd/sssd.conf', raiseonerr=False)

        # Run getent passwd
        usr_cmd = multihost.client[0].run_command(
            f'getent passwd {user}', raiseonerr=False)

        # Try ssh after socket activation is configured
        # Result does not matter we just need to trigger the PAM stack
        with pytest.raises(Exception):
            check_login_client(multihost, user, 'Secret1234')

        # Download sssd pam log
        log_str = multihost.client[0].get_file_contents(
            f"/var/log/sssd/sssd_{domain_name}.log").decode('utf-8')

        # Disable socket activation
        multihost.client[0].run_command(
            "systemctl disable sssd-nss.socket sssd-pam.socket",
            raiseonerr=False)

        # Evaluate test results
        assert usr_cmd.returncode == 0, f"User {user} was not found."
        assert "CONFDB: /var/lib/sss/db/config.ldb" in log_str
        assert "Unable to open tdb '/var/lib/sss/db/config.ldb': " \
               "Permission denied" not in log_str
        assert "Failed to connect to '/var/lib/sss/db/config.ldb'" \
            not in log_str
        assert "The confdb initialization failed" not in log_str

    @staticmethod
    @pytest.mark.tier1
    def test_0009_dbus_method_find_usrby_attr(multihost, backupsssdconf, ldap_posix_usergroup):
        """
        :title: D-Bus method to find user by attributes
        :id: ee3437ff-572e-472e-8f55-0d7d8134266c
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2128840
        :customerscenario: true
        :setup:
          1. In sssd.conf
             set ldap_user_extra_attrs = sn:sn in domain section of sssd.conf
             set user_attributes = +sn in [ifp] section
          2. Create 10 users with 'sn' attribute containing pattern foo
          3. Create a user who does not have sn=foo pattern in it's sn attribute
        :steps:
          1. Restart sssd with clean cache
          2. Fetch users with attribute 'sn:foo*' with dbus-send command
          3. Confirm dbus-send command output has all users with foo*
          4. Confirm dbus-send command output does not contain user
             who donot have sn=foo* matching pattern
        :expectedresults:
          1. SSSD should start successfully
          2. Command should be completed successfully
          3. Expected users with sn:foo* are returned
          4. Expected users without sn:foo* are not returned
        :teardown:
          1. Delete users and groups created for test
          2. Restore sssd.conf
        """
        usr = ldap_posix_usergroup
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        domain_params = {'ldap_search_base': ds_suffix,
                         'auth_provider': 'ldap',
                         'id_provider': 'ldap',
                         'ldap_uri': f'ldaps://{multihost.master[0].sys_hostname}',
                         'ldap_tls_cacert': '/etc/openldap/cacerts/cacert.pem',
                         'use_fully_qualified_names': 'True',
                         'debug_level': '9',
                         'ldap_user_extra_attrs': 'sn:sn'}
        client.sssd_conf(f'domain/{domain_name}', domain_params)
        client.sssd_conf('ifp', {'user_attributes': '+sn'}, action='add')
        client.clear_sssd_cache()
        dbuscmd = 'dbus-send --system --print-reply --dest=org.freedesktop.sssd.infopipe '\
                  '/org/freedesktop/sssd/infopipe/Users org.freedesktop.sssd.infopipe.Users.ListByAttr '\
                  '"string:sn" "string:foo*" "uint32:0"'
        cmd = multihost.client[0].run_command(dbuscmd, raiseonerr=False)
        for i in range(10):
            cmd2 = multihost.client[0].run_command(f'id -u foo{i}@{domain_name}', raiseonerr=False)
            assert cmd2.stdout_text.strip('\n') in cmd.stdout_text, 'dbus is not fetching expected users'
        cmd1 = multihost.client[0].run_command(f'id -u {usr}@{domain_name}', raiseonerr=False)
        assert cmd1.stdout_text.strip('\n') not in cmd.stdout_text, 'dbus is fetching unwanted user'

    @staticmethod
    @pytest.mark.tier1_4
    def test_bz822236(multihost, backupsssdconf):
        """
        :title: Netgroups do not honor entry cache nowait percentage
        :id: dda33ba4-ef10-11ed-a27d-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=822236
        :setup:
            1. Retrieve the name of the network interface that is currently connected using
                the 'nmcli' command, and saves it to the 'intf' variable.
            2. Sets up an LDAP connection to the LDAP server using the LdapOperations class.
            3. Creates a new organizational unit named "Netgroup" under the base DN
                specified in the 'ds_suffix' variable.
            4. Creates a new LDAP entry for a netgroup named "netgrp_nowait"
                under the "Netgroup" organizational unit.
            5. Use the sssdTools class to update the configuration file for the
                'nss' and 'domain/example1' sections of the SSSD service.
            7. Clear the SSSD cache using the 'clear_sssd_cache' method of the sssdTools class.
            8. Delete the contents of the '/var/log/sssd/sssd_nss.log' file
            9. Add a 50ms delay to the network interface using the 'tc' command.
        :steps:
            1. Measures the response time for the 'getent netgroup netgrp_nowait'
                command and saves it to the 'res_time' variable.
            2. Run a loop that repeats the 'getent netgroup netgrp_nowait' command 4 times
                and checks if the response time is less than to the initial response time.
            3. Wait for 15 seconds before deleting the contents of the '/var/log/sssd/sssd_nss.log' file again.
            4. Remove the network delay added in step 9 using the 'tc' command.
        :expectedresults:
            1. res_time variable will have the response time
            2. Response time is less than to the initial response time
            3. Wait for 15 seconds
            4. Network delay should be removed
        """
        client = multihost.client[0]
        multihost.client[0].run_command("modprobe sch_netem")
        log_nss = '/var/log/sssd/sssd_nss.log'
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        ldap_inst.org_unit("Netgroup", ds_suffix)
        user_dn = f'cn=netgrp_nowait,ou=Netgroup,{ds_suffix}'
        user_info = {'cn': 'netgrp_nowait'.encode('utf-8'),
                     'objectClass': ['nisNetgroup'.encode('utf-8'),
                                     'top'.encode('utf-8')],
                     'nisNetgroupTriple': '(host1,kau10,example.com)'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        tools = sssdTools(multihost.client[0])
        tools.sssd_conf("nss", {'filter_groups': 'root',
                                'filter_users': 'root',
                                'debug_level': '9',
                                'entry_cache_nowait_percentage': '50'}, action='update')
        tools.sssd_conf("domain/example1",
                        {'entry_cache_timeout': '30',
                         'ldap_netgroup_search_base': f"ou=Netgroup,{ds_suffix}"}, action='update')
        tools.clear_sssd_cache()
        client.run_command(f"> {log_nss}")
        intf = [s for s in client.run_command("nmcli").stdout_text.split('\n')
                if re.search(r'\b' + "connected to" + r'\b', s)][0].split(":")[0]
        client.run_command(f"tc qdisc add dev {intf} root netem delay 50ms")
        start = D_T.now()
        client.run_command("getent netgroup netgrp_nowait")
        end = D_T.now()
        res_time = end - start
        time.sleep(16)
        time_diff = []
        find_logs_results = []
        for _ in range(4):
            start = D_T.now()
            client.run_command("getent netgroup netgrp_nowait")
            end = D_T.now()
            loop_response = end - start
            time_diff.append(loop_response < res_time)
            time.sleep(3)
            find_logs_results.append(find_logs(multihost,
                                               log_nss,
                                               "Performing midpoint cache "
                                               "update of [netgrp_nowait@example1]"))
            client.run_command(f"> {log_nss}")
            time.sleep(15)
        client.run_command(f"tc qdisc del dev {intf} root")
        ldap_inst.del_dn(user_dn)
        ldap_inst.del_dn(f"ou=Netgroup,{ds_suffix}")
        assert all(find_logs_results), "Searched string not found in the logs"
        assert all(time_diff), "Test failed as the cache response time is higher."

    @staticmethod
    @pytest.mark.tier1_2
    def test_0017_filesldap(multihost, backupsssdconf, setup_sssd_krb):
        """
        :title: sssd-be tends to run out of system resources,
            hitting the maximum number of open files
        :id: ad4a4a32-05d7-11ee-9caa-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2195919
        :setup:
            1. Replace /usr/libexec/sssd/krb5_child with a shell script like
                #!/bin/bash
                sleep 10
        :steps:
            1. Login with ldap user and count no of file descriptors before modifying krb5_child
            2. Trigger authentication requests 'sssctl user-checks -a auth
            3. Count no of file descriptors after modifying krb5_child
        :expectedresults:
            1. Login should success
            2. Authentication requests will fail
            3. No of file descriptors should be same as before and after modifying krb5_child
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        tools.clear_sssd_cache()
        user = f'foo1@{ds_instance_name}'
        check_login_client(multihost, user, "Secret123")
        time.sleep(3)
        client.run_command("cp -af /usr/libexec/sssd/krb5_child /usr/libexec/sssd/krb5_child_bkp")
        new_child = textwrap.dedent("""#!/bin/bash\nsleep 10 \n""")
        client.put_file_contents('/usr/libexec/sssd/krb5_child', new_child)
        client.run_command(f"echo 'Secret123'|sssctl user-checks -a auth foo1@{ds_instance_name}")
        time.sleep(3)
        client.run_command('ls -al /proc/$(pidof sssd_be)/fd > /tmp/before_count')
        time.sleep(2)
        n_log_bfr = count_pattern_logs(multihost, "/tmp/before_count", "pipe:")
        for i in range(10):
            client.run_command(f"echo 'Secret123'|sssctl user-checks -a auth foo{i}@{ds_instance_name}")
            time.sleep(3)
            client.run_command(f'ls -al /proc/$(pidof sssd_be)/fd > /tmp/after_count{i}')
        time.sleep(2)
        n_log_afr = []
        for i in range(10):
            n_log_afr.append(count_pattern_logs(multihost, f"/tmp/after_count{i}", "pipe:"))
        client.run_command("cp -a /usr/libexec/sssd/krb5_child_bkp /usr/libexec/sssd/krb5_child")
        for f_file in ["/tmp/before_count", "/usr/libexec/sssd/krb5_child_bkp"]:
            client_remove_file(multihost, f_file)
        for i in range(10):
            client_remove_file(multihost, f"/tmp/after_count{i}")
        for n_n in n_log_afr:
            assert abs(n_log_bfr - n_n) <= 2



""" Automation of password Policy suite
:requirement: ldap_id_ldap_auth
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import time
import pytest
import ldap
import os
import subprocess
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_instance_name, ds_rootpw, ds_rootdn
from sssd.testlib.common.ssh2_python import check_login_client


def find_logs(multihost, log_name, string_name):
    """This function will find strings in a log file
    log_name: Absolute path of log where the search will happen.
    string_name: String to search in the log file.
    """
    log_str = multihost.client[0].get_file_contents(log_name).decode('utf-8')
    assert string_name in log_str


@pytest.fixture(scope='function')
def common_sssd_setup(multihost):
    """
    This is common sssd setup used in this test suit.
    """
    tools = sssdTools(multihost.client[0])
    tools.sssd_conf("nss", {'filter_groups': 'root',
                            'filter_users': 'root',
                            'debug_level': '9'}, action='update')
    ldap_params = {'use_fully_qualified_names': False}
    tools.sssd_conf(f'domain/{ds_instance_name}', ldap_params)
    tools.clear_sssd_cache()


def ldap_modify_ds(multihost, ldap_mode, dn_dn, element, value):
    """This funciton will do modification on master elements
    ldap_mode: Mode of ldap operation.
    dn_dn: Dn to operate.
    element: Attribute of Dn which needs to change
    value: New Attribute value.
    """
    ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    modify_gid = [(ldap_mode, element, value)]
    ldap_inst.modify_ldap(dn_dn, modify_gid)


@pytest.fixture(scope='class')
def create_users_groups(multihost, request):
    """ Create user and restore """
    client = multihost.client[0]
    ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    user_dn = 'uid=ppuser1,ou=People,dc=example,dc=test'
    user_info = {'cn': 'ppuser1'.encode('utf-8'),
                 'sn': 'ppuser1'.encode('utf-8'),
                 'uid': 'ppuser1'.encode('utf-8'),
                 'loginshell': '/bin/bash'.encode('utf-8'),
                 'uidNumber': '23579'.encode('utf-8'),
                 'gidNumber': '23579'.encode('utf-8'),
                 'objectClass': ['top'.encode('utf-8'),
                                 'inetuser'.encode('utf-8'),
                                 'posixAccount'.encode('utf-8'),
                                 'person'.encode('utf-8'),
                                 'inetorgperson'.encode('utf-8')],
                 'homeDirectory': '/home/ppuser1'.encode('utf-8'),
                 'userPassword': 'Secret123'.encode('utf-8')}
    ldap_inst.add_entry(user_info, user_dn)
    ldap_modify_ds(multihost, ldap.MOD_REPLACE, 'cn=config', 'nsslapd-allowed-to-delete-attrs',
                   [b'passwordExp passwordMaxAge passwordWarning passwordGraceLimit passwordMustChange'])
    file_location = '/script/change_user_password_while_expired.sh'
    client.transport.put_file(os.path.dirname(os.path.abspath(__file__))
                              + file_location, '/tmp/change_user_password_while_expired.sh')

    def restore():
        "Delete user after test finish"
        ldap_inst.del_dn(user_dn)
    request.addfinalizer(restore)


@pytest.mark.tier1_4
@pytest.mark.usefixtures('setup_sssd',
                         'create_users_groups')
@pytest.mark.passwordcheck
class TestPasswordPolicy():
    """
    This is test case class for ldap Password Policy suite
    """
    @staticmethod
    def test_bz748856(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: Set passwordMaxAge to 24 hours bz748856
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=748856
        :id: b0200b74-9675-11ed-b874-845cf3eff344
        :setup:
            1. setup 389-ds w/server-side pw controls
            2. create an account and set the password
            3. clear secure log
            4. clear sssd logs and cache
        :steps:
            1. The user is able to authenticate using SSH password by check_login_client function.
            2. Searches for specific log messages in two log files (/var/log/secure and
                /var/log/sssd/sssd_{ds_instance_name}.log)
                using the find_logs() function, and raises an assertion error if
                the specified log message is not found in the log file.
        :expectedresults:
            1. Authentication should succeed
            2. Corresponding logs should be generated
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        # setup 389-ds w/server-side pw controls
        cn_config = 'cn=config'
        for element, value in [('passwordExp', [b'on']),
                               ('passwordMaxAge', [b'86400']),
                               ('passwordWarning', [b'86400'])]:
            ldap_modify_ds(multihost, ldap.MOD_REPLACE, cn_config, element, value)
        # set the password
        user_dn = 'uid=ppuser1,ou=People,dc=example,dc=test'
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, user_dn, 'userPassword', [b'Secret123'])
        client.run_command('> /var/log/secure')
        tools.clear_sssd_cache()
        check_login_client(multihost, 'ppuser1', 'Secret123')
        time.sleep(3)
        file_scure = '/var/log/secure'
        file_ssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        find_logs(multihost, file_scure, "Your password will expire in ")
        find_logs(multihost, file_ssd, "Server returned control [1.3.6.1.4.1.42.2.27.8.5.1]")
        find_logs(multihost, file_ssd, "Password will expire in [86")
        with pytest.raises(AssertionError):
            find_logs(multihost, file_ssd,
                      "Server does not support the requested "
                      "control [1.3.6.1.4.1.42.2.27.8.5.1]")

    @staticmethod
    def test_maxage(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: "Set passwordMaxAge to 1"
        :id: fc68293e-9676-11ed-9808-845cf3eff344
        :setup:
            1. setup 389-ds w/server-side passwordMaxAge to 1.
            2. sets the password for a user.
            3. clears the SSSD cache for the selected client.
        :steps:
            1. Runs a shell script (/tmp/change_user_password_while_expired.sh) on the selected
                client to simulate a password change attempt after the user's password has expired.
            2. Searches for specific log messages in two log files (/var/log/secure and
                /var/log/sssd/sssd_{ds_instance_name}.log) using the find_logs() function to verify that
                the expected log messages have been generated.
            3. Sets the passwordExp configuration in the cn=config section of the
                389-ds directory server to off using the ldap_modify_ds() function.
        :expectedresults:
            1. User should not be able to reset expired password
            2. Corresponding logs should be generated
            3. PasswordExp configuration in the cn=config section should success
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        cn_config = 'cn=config'
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, cn_config, 'passwordMaxAge', [b'1'])
        user_dn = 'uid=ppuser1,ou=People,dc=example,dc=test'
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, user_dn, 'userPassword', [b'Secret123'])
        client.run_command('> /var/log/secure')
        tools.clear_sssd_cache()
        client.run_command('sh /tmp/change_user_password_while_expired.sh', raiseonerr=False)
        time.sleep(3)
        file_scure = '/var/log/secure'
        file_ssd = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        log_str = multihost.client[0].get_file_contents(file_scure).decode('utf-8')
        assert "Password expired. Change your password now." in log_str or \
               "Failed password for ppuser1" in log_str
        find_logs(multihost, file_ssd, "Server returned control [1.3.6.1.4.1.42.2.27.8.5.1]")
        log_str = multihost.client[0].get_file_contents(file_ssd).decode('utf-8')
        assert "Password expired user must set a new password" in log_str or \
               "Password expired, grace logins exhausted." in log_str
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, cn_config, 'passwordExp', [b'off'])
        time.sleep(3)
        check_login_client(multihost, 'ppuser1', 'Secret123')

    @staticmethod
    def test_bz954323(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: Display last grace login bz954323
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=954323
        :id: f6e2511a-9676-11ed-a446-845cf3eff344
        :setup:
            1. Enable password policy in Directory Server.
            2. Set the number of grace login in password policy to 3
            3. Expire a user password ( by setting password Expiration Time)
            4. Clear the secure log
        :steps:
            1. Calling check_login_client to authenticate
                the user 'ppuser1' with a password on the client.
            2. Calling find_logs to search for a specific message in a log file on the server.
            3. Repeating steps 1-2 two more times, with different log search messages each time.
            4. Calling ldap_modify_ds two more times with different arguments to
                delete the configuration settings modified earlier.
        :expectedresults:
            1. Authentication should succeed
            2. Corresponding logs should be generated
            3. Corresponding logs should be generated
            4. Configuration settings should changed to modified earlier
        """
        client = multihost.client[0]
        cn_config = 'cn=config'
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, cn_config, 'passwordExp', [b'on'])
        ldap_modify_ds(multihost, ldap.MOD_ADD, cn_config, 'passwordGraceLimit', [b'3'])
        client.run_command("> /var/log/secure")
        time.sleep(3)
        check_login_client(multihost, 'ppuser1', 'Secret123')
        time.sleep(3)
        find_logs(multihost, "/var/log/secure", "You have 2 grace login(s) remaining")
        client.run_command("> /var/log/secure")
        check_login_client(multihost, 'ppuser1', 'Secret123')
        time.sleep(3)
        find_logs(multihost, "/var/log/secure", "You have 1 grace login(s) remaining")
        client.run_command("> /var/log/secure")
        check_login_client(multihost, 'ppuser1', 'Secret123')
        time.sleep(3)
        find_logs(multihost, "/var/log/secure", "You have 0 grace login(s) remaining")
        for element, value in [('passwordGraceLimit', []),
                               ('passwordMaxAge', [])]:
            ldap_modify_ds(multihost, ldap.MOD_DELETE, cn_config, element, value)

    @staticmethod
    def test_bz1146198_bz1144011(multihost, backupsssdconf, common_sssd_setup):
        """
        :title: Password expiration policies are not being enforced by SSSD bz1146198 bz1144011
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1146198
                   https://bugzilla.redhat.com/show_bug.cgi?id=1144011
        :id: f6e2511a-9676-11ed-a446-845cf3eff344
        :setup:
            1. Add passwordMustChange to config.
            2. Set the password for the user to 'Secret123'
            3. Clear secure logs
        :steps:
            1. Run a script to simulate an expired password.
            2. Calling find_logs to search for specific messages in log files on the server.
            3. Calling check_login_client to authenticate the user with a new password.
            4. Calling ldap_modify_ds three times to delete some configuration settings related to
                password expiration for the directory
        :expectedresults:
            1. User should be forced to change the password.
            2. Corresponding logs should be generated
            3. User authentication should success
            4. New setting related to password change should success
        """
        client = multihost.client[0]
        cn_config = 'cn=config'
        ldap_modify_ds(multihost, ldap.MOD_ADD, cn_config, 'passwordMustChange', [b'on'])
        user_dn = 'uid=ppuser1,ou=People,dc=example,dc=test'
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, user_dn, 'userPassword', [b'Secret123'])
        client.run_command("> /var/log/secure")
        client.run_command(f"> '/var/log/sssd/sssd_{ds_instance_name}.log'")
        client.run_command("sh /tmp/change_user_password_while_expired.sh")
        time.sleep(3)
        find_logs(multihost, "/var/log/secure", "Password expired. Change your password now")
        find_logs(multihost, f'/var/log/sssd/sssd_{ds_instance_name}.log',
                  "Password was reset. User must set a new password")
        time.sleep(3)
        check_login_client(multihost, 'ppuser1', 'NewPass_123')
        for element, value in [('passwordMustChange', []),
                               ('passwordWarning', []),
                               ('passwordExp', [b'on'])]:
            ldap_modify_ds(multihost, ldap.MOD_DELETE, cn_config, element, value)
        ldap_modify_ds(multihost, ldap.MOD_REPLACE, user_dn, 'userPassword', [b'Secret123'])



"""
Automount Test Cases

:requirement: autofs
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.roles.nfs import NFS
from sssd_test_framework.topology import KnownTopologyGroup


@pytest.mark.importance("critical")
@pytest.mark.ticket(gh=6739)
@pytest.mark.parametrize("cache_first", [False, True])
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_autofs__cache_first_set_to_true(
    client: Client, nfs: NFS, provider: GenericProvider, cache_first: bool, sssd_service_user: str
):
    """
    :title: Autofs works correctly with any cache_first value
    :setup:
        1. Create NFS export
        2. Create auto.master map
        3. Create auto.export map
        4. Add /export (auto.export) key to auto.master
        5. Add "NFS export" key as "export" to auto.export
        6. Enable autofs responder
        7. Set [autofs]/cache_first = $cache_first
        8. Start SSSD
        9. Reload autofs daemon
    :steps:
        1. Access /export/export
        2. Dump automount maps "automount -m"
    :expectedresults:
        1. Directory can be accessed and it is correctly mounted to the NFS share
        2. /export contains auto.export map and "export" key
    :customerscenario: False
    """
    nfs_export = nfs.export("export").add()
    auto_master = provider.automount.map("auto.master").add()
    auto_export = provider.automount.map("auto.export").add()
    auto_master.key("/export").add(info=auto_export)
    key = auto_export.key("export").add(info=nfs_export)

    # Start SSSD
    client.sssd.common.autofs()
    client.sssd.autofs["cache_first"] = str(cache_first)
    client.sssd.start(service_user=sssd_service_user)

    # Reload automounter in order fetch updated maps
    client.automount.reload()

    # Check that we can mount the exported directory
    assert client.automount.mount("/export/export", nfs_export), "Unable to mount /export/export!"

    # Check that the maps are correctly fetched
    assert client.automount.dumpmaps() == {
        "/export": {"map": "auto.export", "keys": [str(key)]},
    }, "Automount maps do not match!"


@pytest.mark.importance("medium")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_autofs__propagate_offline_status_for_a_single_domain(client: Client, provider: GenericProvider):
    """
    :title: Autofs propagates offline status if a domain is offline
    :setup:
        1. Block traffic to the provider
        2. Enable autofs responder
        3. Start SSSD
        4. Reload autofs daemon
    :steps:
        1. Read autofs responder logs
    :expectedresults:
        1. cache_req returns "SSSD is offline" when data provider is offline for auto.master search
    :customerscenario: False
    """
    # Render the provider offline
    client.firewall.outbound.reject_host(provider)

    # Start SSSD
    client.sssd.common.autofs()
    client.sssd.start()

    # Reload automounter in order fetch updated maps
    client.automount.reload()

    # Check that offline status was returned from cache req
    log = client.fs.read(client.sssd.logs.autofs).splitlines()
    offline_status_propagated = False
    for index, line in enumerate(log):
        if "cache_req_process_result" in line and "Finished: Error" in line and "SSSD is offline" in line:
            if "Object [auto.master] was not found in cache" in log[index - 1]:
                offline_status_propagated = True
                break

    assert offline_status_propagated, "Offline status not propagated!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_autofs__propagate_offline_status_for_multiple_domains(client: Client):
    """
    :title: Autofs propagates offline status if a domain is offline in multi domain environment
    :setup:
        1. Create two fake LDAP domains that will be offline (the provider is online but does not have autofs maps)
        2. Enable autofs responder
        3. Start SSSD
        4. Reload autofs daemon
    :steps:
        1. Read autofs responder logs
    :expectedresults:
        1. cache_req returns "SSSD is offline" when data provider is offline for auto.master search
    :customerscenario: False
    """
    # Create fake domains, these will be offline
    client.sssd.dom("fake1").update(
        enabled="true",
        id_provider="ldap",
        ldap_uri="ldap://fake1.test",
    )

    client.sssd.dom("fake2").update(
        enabled="true",
        id_provider="ldap",
        ldap_uri="ldap://fake2.test",
    )

    # Start SSSD
    client.sssd.common.autofs()
    client.sssd.start()

    # Reload automounter in order fetch updated maps
    client.automount.reload()

    # Check that offline status was returned from cache req
    log = client.fs.read(client.sssd.logs.autofs).splitlines()
    offline_status_propagated = False
    for index, line in enumerate(log):
        if "cache_req_process_result" in line and "Finished: Error" in line and "SSSD is offline" in line:
            if "Object [auto.master] was not found in cache" in log[index - 1]:
                offline_status_propagated = True
                break

    assert offline_status_propagated, "Offline status not propagated!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_autofs__works_with_some_offline_domains(client: Client, nfs: NFS, provider: GenericProvider):
    """
    :title: Autofs works if some domain is offline in multi domain environment
    :setup:
        1. Create NFS export
        2. Create auto.master map
        3. Create auto.export map
        4. Add /export (auto.export) key to auto.master
        5. Add "NFS export" key as "export" to auto.export
        6. Create two fake LDAP domains that will be offline (the provider is online)
        7. Enable autofs responder
        8. Start SSSD
        9. Reload autofs daemon
    :steps:
        1. Access /export/export
        2. Dump automount maps "automount -m"
    :expectedresults:
        1. Directory can be accessed and it is correctly mounted to the NFS share
        2. /export contains auto.export map and "export" key
    :customerscenario: False
    """

    # Create autofs maps
    nfs_export = nfs.export("export").add()
    auto_master = provider.automount.map("auto.master").add()
    auto_export = provider.automount.map("auto.export").add()
    auto_master.key("/export").add(info=auto_export)
    key = auto_export.key("export").add(info=nfs_export)

    # Create fake domains, these will be offline
    client.sssd.dom("fake1").update(
        enabled="true",
        id_provider="ldap",
        ldap_uri="ldap://fake1.test",
    )

    client.sssd.dom("fake2").update(
        enabled="true",
        id_provider="ldap",
        ldap_uri="ldap://fake2.test",
    )

    # Start SSSD
    client.sssd.sssd["domain_resolution_order"] = f"fake1, fake2, {client.sssd.default_domain}"
    client.sssd.common.autofs()
    client.sssd.start()

    # Reload automounter in order fetch updated maps
    client.automount.reload()

    # Check that we can mount the exported directory
    assert client.automount.mount("/export/export", nfs_export), "Unable to mount /export/export!"

    # Check that the maps are correctly fetched
    assert client.automount.dumpmaps() == {
        "/export": {"map": "auto.export", "keys": [str(key)]},
    }, "Automount maps do not match!"
